from SolMuseum.pde.gas.util import *

__all__ = ['rupture_ngs_pipe_cdm', 'leakage_ngs_pipe_cdm']


def rupture_ngs_pipe_cdm(p: Var,
                         q: Var,
                         lam,
                         va,
                         D,
                         S,
                         dx,
                         dt,
                         M,
                         pipe_name: str,
                         idx_leak):
    r"""
    This function model and discretize natural gas equations

        .. math::
            \pdv{p}{t}+\frac{c^2}{S}\pdv{q}{x}=&0,\\
            \pdv{q}{t}+S\pdv{p}{x}+\frac{\lambda c^2q|q|}{2DSp}=&0

    Parameters
    ==========

    p : Var
        The pressure w.r.t. x and t

    q : Var
        The mass flow w.r.t. x and t

    lam : Param or Number
        The friction $\lambda$

    va : Param or Number
        The speed of sound

    D : Param or Number
        The pipe diameter

    S : Param or Number
        The cross-section area

    dx : Param or Number
        The spatial difference step size

    M : Param or Number
        The friction $\lambda$

    pipe_name : str
        The name of the pipe

    method : str

        'cha' - The method of characteristics

            .. math ::
                p_i^{j+1}-p_{i-1}^j+\frac{c}{S}\left(q_i^{j+1}-q_{i-1}^j\right)+\frac{\lambda c^2 \Delta x}{4 D S^2} \frac{\left(q_i^{j+1}+q_{i-1}^j\right)\left|q_i^{j+1}+q_{i-1}^j\right|}{p_i^{j+1}+p_{i-1}^j}&=0,\quad 1\leq i\leq M,\\
                p_{i+1}^j-p_i^{j+1}+\frac{c}{S}\left(q_i^{j+1}-q_{i+1}^j\right)+\frac{\lambda c^2 \Delta x}{4 D S^2} \frac{\left(q_i^{j+1}+q_{i+1}^j\right)\left|q_i^{j+1}+q_{i+1}^j\right|}{p_i^{j+1}+p_{i+1}^j}&=0,\quad 0\leq i\leq M-1.

        'weno' - The WENO semi-discretization

            .. math ::
                \pdv{u_j}{t}=-\frac{1}{\Delta x}\qty(\hat{f}_{j+1/2}-\hat{f}_{j-1/2})+S(u_j)

            where $\hat{f}_{j+1/2}$ and $\hat{f}_{j-1/2}$ are reconstructed by the weighted essentially non-oscillatory
            scheme [1]_.

    Returns
    =======

    artifact : dict()
        The dict of derived equations and variables.

    References
    ==========

    .. [1] C.-W. Shu, “Essentially non-oscillatory and weighted essentially non- oscillatory schemes for hyperbolic conservation laws,” in Advanced Numerical Approximation of Nonlinear Hyperbolic Equations: Lectures given at the 2nd Session of the Centro Internazionale Matematico Estivo (C.I.M.E.) held in Cetraro, Italy, June 23–28, 1997, A. Quarteroni, Ed. Berlin, Heidelberg: Springer Berlin Heidelberg, 1998, pp. 325–432.

    """

    if not is_integer(M):
        raise TypeError(f'M is {type(M)} instead of integer')

    if not is_number(dx):
        raise TypeError(f'dx is {type(dx)} instead of number')

    artifact = dict()

    qleak1 = Var(f'q_{pipe_name}_leak1', init=q[idx_leak - 1])
    qleak2 = Var(f'q_{pipe_name}_leak2', init=q[idx_leak + 1])
    qleak1_0 = AliasVar(qleak1.name, init=qleak1)
    qleak2_0 = AliasVar(qleak2.name, init=qleak2)

    artifact[qleak1.name] = qleak1
    artifact[qleak2.name] = qleak2
    artifact[qleak1_0.name] = qleak1_0
    artifact[qleak2_0.name] = qleak2_0

    p0 = AliasVar(p.name, init=p)
    q0 = AliasVar(q.name, init=q)
    artifact[p0.name] = p0
    artifact[q0.name] = q0

    # 1<=i<=idx_leak-1
    rhs = cdm(p[1:idx_leak], p[0:idx_leak - 1], p0[1:idx_leak], p0[0:idx_leak - 1],
              q[1:idx_leak], q[0:idx_leak - 1], q0[1:idx_leak], q0[0:idx_leak - 1],
              'p', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_p_0'] = Eqn(f'cdm_{pipe_name}_p_0',
                                           rhs)

    rhs = cdm(p[1:idx_leak], p[0:idx_leak - 1], p0[1:idx_leak], p0[0:idx_leak - 1],
              q[1:idx_leak], q[0:idx_leak - 1], q0[1:idx_leak], q0[0:idx_leak - 1],
              'q', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_q_0'] = Eqn(f'cdm_{pipe_name}_q_0',
                                           rhs)

    # i=idx_leak
    rhs = cdm(p[idx_leak], p[idx_leak - 1], p0[idx_leak], p0[idx_leak - 1],
              qleak1, q[idx_leak - 1], qleak1_0, q0[idx_leak - 1],
              'p', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_p_1'] = Eqn(f'cdm_{pipe_name}_p_1', rhs)

    rhs = cdm(p[idx_leak], p[idx_leak - 1], p0[idx_leak], p0[idx_leak - 1],
              qleak1, q[idx_leak - 1], qleak1_0, q0[idx_leak - 1],
              'q', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_q_1'] = Eqn(f'cdm_{pipe_name}_q_1', rhs)

    # i=idx_leak+1
    rhs = cdm(p[idx_leak + 1], p[idx_leak], p0[idx_leak + 1], p0[idx_leak],
              q[idx_leak + 1], qleak2, q0[idx_leak + 1], qleak2_0,
              'p', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_p_2'] = Eqn(f'cdm_{pipe_name}_p_2',
                                           rhs)

    rhs = cdm(p[idx_leak + 1], p[idx_leak], p0[idx_leak + 1], p0[idx_leak],
              q[idx_leak + 1], qleak2, q0[idx_leak + 1], qleak2_0,
              'q', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_q_2'] = Eqn(f'cdm_{pipe_name}_q_2',
                                           rhs)

    # idx_leak+2<=i<=M
    rhs = cdm(p[idx_leak + 2:M + 1], p[idx_leak + 1:M], p0[idx_leak + 2:M + 1], p0[idx_leak + 1:M],
              q[idx_leak + 2:M + 1], q[idx_leak + 1:M], q0[idx_leak + 2:M + 1], q0[idx_leak + 1:M],
              'p', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_p_3'] = Eqn(f'cdm_{pipe_name}_p_3',
                                           rhs)

    rhs = cdm(p[idx_leak + 2:M + 1], p[idx_leak + 1:M], p0[idx_leak + 2:M + 1], p0[idx_leak + 1:M],
              q[idx_leak + 2:M + 1], q[idx_leak + 1:M], q0[idx_leak + 2:M + 1], q0[idx_leak + 1:M],
              'q', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_q_3'] = Eqn(f'cdm_{pipe_name}_q_3',
                                           rhs)

    pb0 = p.value[idx_leak]
    pa_t = [pb0, pb0, pb0, pb0]
    tseries = [0, 300, 300 + 10, 10 * 3600]
    pa = TimeSeriesParam('pa',
                         v_series=pa_t,
                         time_series=tseries)
    artifact['pa'] = pa

    artifact[f'cdm_{pipe_name}_bd1'] = Eqn(f'cdm_{pipe_name}_bd1',
                                           p[idx_leak] - pa)
    artifact[f'cdm_{pipe_name}_bd2'] = Eqn(f'cdm_{pipe_name}_bd2',
                                           qleak1 - qleak2 - q[idx_leak])
    return artifact


def leakage_ngs_pipe_cdm(p: Var,
                         q: Var,
                         lam,
                         va,
                         D,
                         S,
                         dx,
                         dt,
                         M,
                         pipe_name: str,
                         idx_leak,
                         d):
    if not is_integer(M):
        raise TypeError(f'M is {type(M)} instead of integer')

    if not is_number(dx):
        raise TypeError(f'dx is {type(dx)} instead of number')

    artifact = dict()

    qleak1 = Var(f'q_{pipe_name}_leak1', init=q[idx_leak - 1])
    qleak2 = Var(f'q_{pipe_name}_leak2', init=q[idx_leak + 1])
    qleak1_0 = AliasVar(qleak1.name, init=qleak1)
    qleak2_0 = AliasVar(qleak2.name, init=qleak2)
    is_sonic = Param('is_sonic', value=1)

    artifact[qleak1.name] = qleak1
    artifact[qleak2.name] = qleak2
    artifact[qleak1_0.name] = qleak1_0
    artifact[qleak2_0.name] = qleak2_0
    artifact[is_sonic.name] = is_sonic

    p0 = AliasVar(p.name, init=p)
    q0 = AliasVar(q.name, init=q)
    artifact[p0.name] = p0
    artifact[q0.name] = q0

    # 1<=i<=idx_leak-1
    rhs = cdm(p[1:idx_leak], p[0:idx_leak - 1], p0[1:idx_leak], p0[0:idx_leak - 1],
              q[1:idx_leak], q[0:idx_leak - 1], q0[1:idx_leak], q0[0:idx_leak - 1],
              'p', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_p_0'] = Eqn(f'cdm_{pipe_name}_p_0',
                                           rhs)

    rhs = cdm(p[1:idx_leak], p[0:idx_leak - 1], p0[1:idx_leak], p0[0:idx_leak - 1],
              q[1:idx_leak], q[0:idx_leak - 1], q0[1:idx_leak], q0[0:idx_leak - 1],
              'q', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_q_0'] = Eqn(f'cdm_{pipe_name}_q_0',
                                           rhs)

    # i=idx_leak
    rhs = cdm(p[idx_leak], p[idx_leak - 1], p0[idx_leak], p0[idx_leak - 1],
              qleak1, q[idx_leak - 1], qleak1_0, q0[idx_leak - 1],
              'p', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_p_1'] = Eqn(f'cdm_{pipe_name}_p_1', rhs)

    rhs = cdm(p[idx_leak], p[idx_leak - 1], p0[idx_leak], p0[idx_leak - 1],
              qleak1, q[idx_leak - 1], qleak1_0, q0[idx_leak - 1],
              'q', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_q_1'] = Eqn(f'cdm_{pipe_name}_q_1', rhs)

    # i=idx_leak+1
    rhs = cdm(p[idx_leak + 1], p[idx_leak], p0[idx_leak + 1], p0[idx_leak],
              q[idx_leak + 1], qleak2, q0[idx_leak + 1], qleak2_0,
              'p', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_p_2'] = Eqn(f'cdm_{pipe_name}_p_2',
                                           rhs)

    rhs = cdm(p[idx_leak + 1], p[idx_leak], p0[idx_leak + 1], p0[idx_leak],
              q[idx_leak + 1], qleak2, q0[idx_leak + 1], qleak2_0,
              'q', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_q_2'] = Eqn(f'cdm_{pipe_name}_q_2',
                                           rhs)

    # idx_leak+2<=i<=M
    rhs = cdm(p[idx_leak + 2:M + 1], p[idx_leak + 1:M], p0[idx_leak + 2:M + 1], p0[idx_leak + 1:M],
              q[idx_leak + 2:M + 1], q[idx_leak + 1:M], q0[idx_leak + 2:M + 1], q0[idx_leak + 1:M],
              'p', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_p_3'] = Eqn(f'cdm_{pipe_name}_p_3',
                                           rhs)

    rhs = cdm(p[idx_leak + 2:M + 1], p[idx_leak + 1:M], p0[idx_leak + 2:M + 1], p0[idx_leak + 1:M],
              q[idx_leak + 2:M + 1], q[idx_leak + 1:M], q0[idx_leak + 2:M + 1], q0[idx_leak + 1:M],
              'q', lam, va, S, D, dx, dt)
    artifact[f'cdm_{pipe_name}_q_3'] = Eqn(f'cdm_{pipe_name}_q_3',
                                           rhs)

    P2 = p[idx_leak]
    leak_rate = TimeSeriesParam('leak_rate',
                                v_series=[0, 0],
                                time_series=[0, 3 * 3600])
    artifact['leak_rate'] = leak_rate

    C0 = 0.61
    T0 = 293
    Mass = 17.1e-3
    Z = 1
    R = 8.314
    Hcr = 1.3
    d_leak = d
    Ah = np.pi * (d_leak / 2) ** 2
    Pa = 101e3
    P2cr = Pa * (2 / (Hcr + 1)) ** (-Hcr / (Hcr - 1))

    qjleak = leak_rate * (is_sonic *
                          Ah * P2 *
                          (Mass / (Z * R * T0) * Hcr * (2 / (Hcr + 1)) ** ((Hcr + 1) / (Hcr - 1))) ** (1 / 2)
                          + (1 - is_sonic) *
                          C0 * Ah * P2 *
                          (2 * Mass / (Z * R * T0) * Hcr / (Hcr - 1) * (
                                  (Pa / P2) ** (2 / Hcr) - (Pa / P2) ** ((Hcr + 1) / Hcr))) ** (1 / 2))

    artifact[f'cdm_{pipe_name}_bd1'] = Eqn(f'cdm_{pipe_name}_bd1',
                                           q[idx_leak] - qjleak)
    artifact[f'cdm_{pipe_name}_bd2'] = Eqn(f'cdm_{pipe_name}_bd2',
                                           qleak1 - qleak2 - q[idx_leak])
    return artifact



from SolMuseum.pde.gas.util import *

__all__ = ['ngs_pipe_euler', 'rupture_ngs_pipe_Euler', 'leakage_ngs_pipe_Euler']


def ngs_pipe_euler(p: Var,
                   q: Var,
                   lam,
                   va,
                   D,
                   S,
                   dx,
                   dt,
                   M,
                   pipe_name: str,
                   direction=2):
    r"""
    This function model and discretize natural gas equations

        .. math::
            \pdv{p}{t}+\frac{c^2}{S}\pdv{q}{x}=&0,\\
            \pdv{q}{t}+S\pdv{p}{x}+\frac{\lambda c^2q|q|}{2DSp}=&0

    Parameters
    ==========

    p : Var
        The pressure w.r.t. x and t

    q : Var
        The mass flow w.r.t. x and t

    lam : Param or Number
        The friction $\lambda$

    va : Param or Number
        The speed of sound

    D : Param or Number
        The pipe diameter

    S : Param or Number
        The cross-section area

    dx : Param or Number
        The spatial difference step size

    M : Param or Number
        The friction $\lambda$

    pipe_name : str
        The name of the pipe

    method : str

        'cha' - The method of characteristics

            .. math ::
                p_i^{j+1}-p_{i-1}^j+\frac{c}{S}\left(q_i^{j+1}-q_{i-1}^j\right)+\frac{\lambda c^2 \Delta x}{4 D S^2} \frac{\left(q_i^{j+1}+q_{i-1}^j\right)\left|q_i^{j+1}+q_{i-1}^j\right|}{p_i^{j+1}+p_{i-1}^j}&=0,\quad 1\leq i\leq M,\\
                p_{i+1}^j-p_i^{j+1}+\frac{c}{S}\left(q_i^{j+1}-q_{i+1}^j\right)+\frac{\lambda c^2 \Delta x}{4 D S^2} \frac{\left(q_i^{j+1}+q_{i+1}^j\right)\left|q_i^{j+1}+q_{i+1}^j\right|}{p_i^{j+1}+p_{i+1}^j}&=0,\quad 0\leq i\leq M-1.

        'weno' - The WENO semi-discretization

            .. math ::
                \pdv{u_j}{t}=-\frac{1}{\Delta x}\qty(\hat{f}_{j+1/2}-\hat{f}_{j-1/2})+S(u_j)

            where $\hat{f}_{j+1/2}$ and $\hat{f}_{j-1/2}$ are reconstructed by the weighted essentially non-oscillatory
            scheme [1]_.

    Returns
    =======

    artifact : dict()
        The dict of derived equations and variables.

    References
    ==========

    .. [1] C.-W. Shu, “Essentially non-oscillatory and weighted essentially non- oscillatory schemes for hyperbolic conservation laws,” in Advanced Numerical Approximation of Nonlinear Hyperbolic Equations: Lectures given at the 2nd Session of the Centro Internazionale Matematico Estivo (C.I.M.E.) held in Cetraro, Italy, June 23–28, 1997, A. Quarteroni, Ed. Berlin, Heidelberg: Springer Berlin Heidelberg, 1998, pp. 325–432.

    """

    if not is_integer(M):
        raise TypeError(f'M is {type(M)} instead of integer')

    if not is_number(dx):
        raise TypeError(f'dx is {type(dx)} instead of number')

    artifact = dict()

    match direction:
        case 1:
            p0 = AliasVar(p.name, init=p)
            q0 = AliasVar(q.name, init=q)
            artifact[p0.name] = p0
            artifact[q0.name] = q0
            rhs = (p[1:M + 1] - p0[1:M + 1]) / dt + va ** 2 / S * (q[1:M + 1] - q[0:M]) / dx
            artifact[f'euler{pipe_name}_p'] = Eqn(f'euler{pipe_name}_p', rhs)
            rhs = (q[1:M + 1] - q0[1:M + 1]) / dt + S * (p[1:M + 1] - p[0:M]) / dx + lam * va ** 2 * q[1:M + 1] * Abs(
                q[1:M + 1]) / (2 * D * S * p[1:M + 1])
            artifact[f'euler{pipe_name}_q'] = Eqn(f'euler{pipe_name}_q', rhs)
        case 2:
            p0 = AliasVar(p.name, init=p)
            q0 = AliasVar(q.name, init=q)
            artifact[p0.name] = p0
            artifact[q0.name] = q0
            rhs = (p[0:M] - p0[0:M]) / dt + va ** 2 / S * (q[1:M + 1] - q[0:M]) / dx
            artifact[f'euler{pipe_name}_p'] = Eqn(f'euler{pipe_name}_p', rhs)
            rhs = (q[1:M + 1] - q0[1:M + 1]) / dt + S * (p[1:M + 1] - p[0:M]) / dx + lam * va ** 2 * q[1:M + 1] * Abs(
                q[1:M + 1]) / (2 * D * S * p[1:M + 1])
            artifact[f'euler{pipe_name}_q'] = Eqn(f'euler{pipe_name}_q', rhs)
        case _:
            raise ValueError(f"Direction {direction} not supported!")

    return artifact


def rupture_ngs_pipe_Euler(p: Var,
                           q: Var,
                           lam,
                           va,
                           D,
                           S,
                           dx,
                           dt,
                           M,
                           pipe_name: str,
                           idx_leak):
    if not is_integer(M):
        raise TypeError(f'M is {type(M)} instead of integer')

    if not is_number(dx):
        raise TypeError(f'dx is {type(dx)} instead of number')

    artifact = dict()

    qleak1 = Var(f'q_{pipe_name}_leak1', init=q[idx_leak - 1])
    qleak2 = Var(f'q_{pipe_name}_leak2', init=q[idx_leak + 1])
    qleak1_0 = AliasVar(qleak1.name, init=qleak1)
    qleak2_0 = AliasVar(qleak2.name, init=qleak2)

    artifact[qleak1.name] = qleak1
    artifact[qleak2.name] = qleak2
    artifact[qleak1_0.name] = qleak1_0
    artifact[qleak2_0.name] = qleak2_0

    p0 = AliasVar(p.name, init=p)
    q0 = AliasVar(q.name, init=q)
    artifact[p0.name] = p0
    artifact[q0.name] = q0

    # 1<=i<=idx_leak-1
    rhs = ((p[1:idx_leak] - p0[1:idx_leak]) / dt +
           va ** 2 / S * (q[1:idx_leak] - q[0:idx_leak - 1]) / dx)
    artifact[f'eu_{pipe_name}_p_0'] = Eqn(f'eu_{pipe_name}_p_0',
                                          rhs)

    rhs = ((q[1:idx_leak] - q0[1:idx_leak]) / dt + S * (p[1:idx_leak] - p[0:idx_leak - 1]) / dx +
           lam * va ** 2 * q[1:idx_leak] * Abs(q[1:idx_leak]) / (2 * D * S * p[1:idx_leak]))
    artifact[f'eu_{pipe_name}_q_0'] = Eqn(f'eu_{pipe_name}_q_0',
                                          rhs)

    # i=idx_leak
    rhs = ((p[idx_leak] - p0[idx_leak]) / dt +
           va ** 2 / S * (qleak1 - q[idx_leak - 1]) / dx)
    artifact[f'eu_{pipe_name}_p_1'] = Eqn(f'eu_{pipe_name}_p_1', rhs)

    rhs = ((qleak1 - qleak1_0) / dt + S * (p[idx_leak] - p[idx_leak - 1]) / dx +
           lam * va ** 2 * qleak1 * Abs(qleak1) / (2 * D * S * p[idx_leak]))
    artifact[f'eu_{pipe_name}_q_1'] = Eqn(f'eu_{pipe_name}_q_1', rhs)

    # i=idx_leak+1
    rhs = ((p[idx_leak + 1] - p0[idx_leak + 1]) / dt +
           va ** 2 / S * (q[idx_leak + 1] - qleak2) / dx)
    artifact[f'eu_{pipe_name}_p_2'] = Eqn(f'eu_{pipe_name}_p_2',
                                          rhs)

    rhs = ((q[idx_leak + 1] - q0[idx_leak + 1]) / dt + S * (p[idx_leak + 1] - p[idx_leak]) / dx +
           lam * va ** 2 * q[idx_leak + 1] * Abs(q[idx_leak + 1]) / (2 * D * S * p[idx_leak + 1]))
    artifact[f'eu_{pipe_name}_q_2'] = Eqn(f'eu_{pipe_name}_q_2',
                                          rhs)

    # idx_leak+2<=i<=M
    rhs = ((p[idx_leak + 2:M + 1] - p0[idx_leak + 2:M + 1]) / dt +
           va ** 2 / S * (q[idx_leak + 2:M + 1] - q[idx_leak + 1:M]) / dx)
    artifact[f'eu_{pipe_name}_p_3'] = Eqn(f'eu_{pipe_name}_p_3',
                                          rhs)

    rhs = ((q[idx_leak + 2:M + 1] - q0[idx_leak + 2:M + 1]) / dt + S * (
            p[idx_leak + 2:M + 1] - p[idx_leak + 1:M]) / dx +
           lam * va ** 2 * q[idx_leak + 2:M + 1] * Abs(q[idx_leak + 2:M + 1]) / (
                   2 * D * S * p[idx_leak + 2:M + 1]))
    artifact[f'eu_{pipe_name}_q_3'] = Eqn(f'eu_{pipe_name}_q_3',
                                          rhs)

    pb0 = p.value[idx_leak]
    pa_t = [pb0, pb0, pb0, pb0]
    tseries = [0, 300, 300 + 10, 10 * 3600]
    pa = TimeSeriesParam('pa',
                         v_series=pa_t,
                         time_series=tseries)
    artifact['pa'] = pa

    artifact[f'eu_{pipe_name}_bd1'] = Eqn(f'eu_{pipe_name}_bd1',
                                          p[idx_leak] - pa)
    artifact[f'eu_{pipe_name}_bd2'] = Eqn(f'eu_{pipe_name}_bd2',
                                          qleak1 - qleak2 - q[idx_leak])
    return artifact


def leakage_ngs_pipe_Euler(p: Var,
                           q: Var,
                           lam,
                           va,
                           D,
                           S,
                           dx,
                           dt,
                           M,
                           pipe_name: str,
                           idx_leak,
                           d):
    if not is_integer(M):
        raise TypeError(f'M is {type(M)} instead of integer')

    if not is_number(dx):
        raise TypeError(f'dx is {type(dx)} instead of number')

    artifact = dict()

    qleak1 = Var(f'q_{pipe_name}_leak1', init=q[idx_leak - 1])
    qleak2 = Var(f'q_{pipe_name}_leak2', init=q[idx_leak + 1])
    qleak1_0 = AliasVar(qleak1.name, init=qleak1)
    qleak2_0 = AliasVar(qleak2.name, init=qleak2)
    is_sonic = Param('is_sonic', value=1)

    artifact[qleak1.name] = qleak1
    artifact[qleak2.name] = qleak2
    artifact[qleak1_0.name] = qleak1_0
    artifact[qleak2_0.name] = qleak2_0
    artifact[is_sonic.name] = is_sonic

    p0 = AliasVar(p.name, init=p)
    q0 = AliasVar(q.name, init=q)
    artifact[p0.name] = p0
    artifact[q0.name] = q0

    # 1<=i<=idx_leak-1
    rhs = ((p[1:idx_leak] - p0[1:idx_leak]) / dt +
           va ** 2 / S * (q[1:idx_leak] - q[0:idx_leak - 1]) / dx)
    artifact[f'eu_{pipe_name}_p_0'] = Eqn(f'eu_{pipe_name}_p_0',
                                          rhs)

    rhs = ((q[1:idx_leak] - q0[1:idx_leak]) / dt + S * (p[1:idx_leak] - p[0:idx_leak - 1]) / dx +
           lam * va ** 2 * q[1:idx_leak] * Abs(q[1:idx_leak]) / (2 * D * S * p[1:idx_leak]))
    artifact[f'eu_{pipe_name}_q_0'] = Eqn(f'eu_{pipe_name}_q_0',
                                          rhs)

    # i=idx_leak
    rhs = ((p[idx_leak] - p0[idx_leak]) / dt +
           va ** 2 / S * (qleak1 - q[idx_leak - 1]) / dx)
    artifact[f'eu_{pipe_name}_p_1'] = Eqn(f'eu_{pipe_name}_p_1', rhs)

    rhs = ((qleak1 - qleak1_0) / dt + S * (p[idx_leak] - p[idx_leak - 1]) / dx +
           lam * va ** 2 * qleak1 * Abs(qleak1) / (2 * D * S * p[idx_leak]))
    artifact[f'eu_{pipe_name}_q_1'] = Eqn(f'eu_{pipe_name}_q_1', rhs)

    # i=idx_leak+1
    rhs = ((p[idx_leak + 1] - p0[idx_leak + 1]) / dt +
           va ** 2 / S * (q[idx_leak + 1] - qleak2) / dx)
    artifact[f'eu_{pipe_name}_p_2'] = Eqn(f'eu_{pipe_name}_p_2',
                                          rhs)

    rhs = ((q[idx_leak + 1] - q0[idx_leak + 1]) / dt + S * (p[idx_leak + 1] - p[idx_leak]) / dx +
           lam * va ** 2 * q[idx_leak + 1] * Abs(q[idx_leak + 1]) / (2 * D * S * p[idx_leak + 1]))
    artifact[f'eu_{pipe_name}_q_2'] = Eqn(f'eu_{pipe_name}_q_2',
                                          rhs)

    # idx_leak+2<=i<=M
    rhs = ((p[idx_leak + 2:M + 1] - p0[idx_leak + 2:M + 1]) / dt +
           va ** 2 / S * (q[idx_leak + 2:M + 1] - q[idx_leak + 1:M]) / dx)
    artifact[f'eu_{pipe_name}_p_3'] = Eqn(f'eu_{pipe_name}_p_3',
                                          rhs)

    rhs = ((q[idx_leak + 2:M + 1] - q0[idx_leak + 2:M + 1]) / dt + S * (
            p[idx_leak + 2:M + 1] - p[idx_leak + 1:M]) / dx +
           lam * va ** 2 * q[idx_leak + 2:M + 1] * Abs(q[idx_leak + 2:M + 1]) / (
                   2 * D * S * p[idx_leak + 2:M + 1]))
    artifact[f'eu_{pipe_name}_q_3'] = Eqn(f'eu_{pipe_name}_q_3',
                                          rhs)

    P2 = p[idx_leak]
    leak_rate = TimeSeriesParam('leak_rate',
                                v_series=[0, 0],
                                time_series=[0, 3 * 3600])
    artifact['leak_rate'] = leak_rate

    C0 = 0.61
    T0 = 293
    Mass = 17.1e-3
    Z = 1
    R = 8.314
    Hcr = 1.3
    d_leak = d
    Ah = np.pi * (d_leak / 2) ** 2
    Pa = 101e3
    P2cr = Pa * (2 / (Hcr + 1)) ** (-Hcr / (Hcr - 1))

    qjleak = leak_rate * (is_sonic *
                          Ah * P2 *
                          (Mass / (Z * R * T0) * Hcr * (2 / (Hcr + 1)) ** ((Hcr + 1) / (Hcr - 1))) ** (1 / 2)
                          + (1 - is_sonic) *
                          C0 * Ah * P2 *
                          (2 * Mass / (Z * R * T0) * Hcr / (Hcr - 1) * (
                                  (Pa / P2) ** (2 / Hcr) - (Pa / P2) ** ((Hcr + 1) / Hcr))) ** (1 / 2))

    artifact[f'eu_{pipe_name}_bd1'] = Eqn(f'eu_{pipe_name}_bd1',
                                          q[idx_leak] - qjleak)
    artifact[f'eu_{pipe_name}_bd2'] = Eqn(f'eu_{pipe_name}_bd2',
                                          qleak1 - qleak2 - q[idx_leak])
    return artifact



from numpy import sign, abs
from numba import njit


@njit(cache=True)
def kt1_ode0(pm1, p0, pp1, qm1, q0, qp1, S, va, lam, D, dx):
    return -1 / 2 * S * (-pm1 + pp1) / dx + (1 / 2) * va * (
            -2 * q0 + qm1 + qp1) / dx - 1 / 2 * lam * va ** 2 * q0 * abs(q0) / (D * S * p0)


@njit(cache=True)
def dkt1_ode0dq0(pm1, p0, pp1, qm1, q0, qp1, S, va, lam, D, dx):
    return -va / dx - 1 / 2 * lam * va ** 2 * q0 * sign(q0) / (D * S * p0) - 1 / 2 * lam * va ** 2 * abs(q0) / (
            D * S * p0)


@njit(cache=True)
def dkt1_ode0dqp1(pm1, p0, pp1, qm1, q0, qp1, S, va, lam, D, dx):
    return (1 / 2) * va / dx


@njit(cache=True)
def dkt1_ode0dpp1(pm1, p0, pp1, qm1, q0, qp1, S, va, lam, D, dx):
    return -1 / 2 * S / dx


@njit(cache=True)
def dkt1_ode0dqm1(pm1, p0, pp1, qm1, q0, qp1, S, va, lam, D, dx):
    return (1 / 2) * va / dx


@njit(cache=True)
def dkt1_ode0dpm1(pm1, p0, pp1, qm1, q0, qp1, S, va, lam, D, dx):
    return (1 / 2) * S / dx


@njit(cache=True)
def dkt1_ode0dp0(pm1, p0, pp1, qm1, q0, qp1, S, va, lam, D, dx):
    return (1 / 2) * lam * va ** 2 * q0 * abs(q0) / (D * S * p0 ** 2)



from SolMuseum.pde.gas.util import *


def ngs_pipe(p: Var,
             q: Var,
             lam,
             va,
             D,
             S,
             dx,
             dt,
             M,
             pipe_name: str,
             method='euler'):
    r"""
    This function model and discretize natural gas equations

        .. math::
            \pdv{p}{t}+\frac{c^2}{S}\pdv{q}{x}=&0,\\
            \pdv{q}{t}+S\pdv{p}{x}+\frac{\lambda c^2q|q|}{2DSp}=&0
    where $p$ is the two-dimensional pressure distribution, $p$ is the two-dimensional mass flow distribution,
    $c$ is the sound speed, $S$ is the cross-sectional area, $\lambda$ is the friction coefficient, and $D$ is the pipe
    diameter.

    For semi-discretization method, the boundaries are approximated using characteristics [4]_, [5]_.

    Parameters
    ==========

    p : Var
        The pressure w.r.t. x and t

    q : Var
        The mass flow w.r.t. x and t

    lam : Param or Number
        The friction $\lambda$

    va : Param or Number
        The speed of sound

    D : Param or Number
        The pipe diameter

    S : Param or Number
        The cross-section area

    dx : Param or Number
        The spatial difference step size

    dt : Param or Number
        The temporal step size. `dt` can be set to 0 if one uses the kt2 scheme.

    M : Param or Number
        The friction $\lambda$

    pipe_name : str
        The name of the pipe

    method : str

        'euler' - Default, the euler scheme [3]_

            .. math ::
                \left\{
                \begin{aligned}
                    &\pdv{u}{t}=\frac{u_{i+1}^{j+1}-u_{i+1}^{j}}{\Delta t}\\
                    &\pdv{u}{x}=\frac{u_{i+1}^{j+1}-u_{i}^{j+1}}{\Delta x}\\
                     &u=u_{i+1}^{j+1}
                \end{aligned}
                \right.

        'cdm' - The central difference scheme [3]_


            .. math ::
                \left\{
                \begin{aligned}
                    &\pdv{u}{t}=\frac{u_{i}^{j+1}+u_{i+1}^{j+1}-u_{i}^{j}-u_{i+1}^{j}}{2\Delta t}\\
                    &\pdv{u}{x}=\frac{u_{i+1}^{j+1}+u_{i+1}^{j}-u_{i}^{j+1}-u_{i}^{j}}{2\Delta x}\\
                    &u=\frac{u_{i}^{j+1}+u_{i+1}^{j+1}+u_{i}^{j}+u_{i+1}^{j}}{4}
                \end{aligned}
                \right.

        'cha' - The method of characteristics [3]_

            .. math ::
                p_i^{j+1}-p_{i-1}^j+\frac{c}{S}\left(q_i^{j+1}-q_{i-1}^j\right)+\frac{\lambda c^2 \Delta x}{4 D S^2} \frac{\left(q_i^{j+1}+q_{i-1}^j\right)\left|q_i^{j+1}+q_{i-1}^j\right|}{p_i^{j+1}+p_{i-1}^j}&=0,\quad 1\leq i\leq M,\\
                p_{i+1}^j-p_i^{j+1}+\frac{c}{S}\left(q_i^{j+1}-q_{i+1}^j\right)+\frac{\lambda c^2 \Delta x}{4 D S^2} \frac{\left(q_i^{j+1}+q_{i+1}^j\right)\left|q_i^{j+1}+q_{i+1}^j\right|}{p_i^{j+1}+p_{i+1}^j}&=0,\quad 0\leq i\leq M-1.

        'kt1' - The first order Kurganov-Tadmor semi-discretization scheme

            .. math ::
                \pdv{u_j}{t}=-\frac{1}{\Delta x}\qty(\hat{f}_{j+1/2}-\hat{f}_{j-1/2})+S(u_j)

            where $\hat{f}_{j+1/2}$ and $\hat{f}_{j-1/2}$ are reconstructed by the first order Kurganov-Tadmor scheme [2]_.

        'kt2' - The second order Kurganov-Tadmor semi-discretization scheme

            .. math ::
                \pdv{u_j}{t}=-\frac{1}{\Delta x}\qty(\hat{f}_{j+1/2}-\hat{f}_{j-1/2})+S(u_j)

            where $\hat{f}_{j+1/2}$ and $\hat{f}_{j-1/2}$ are reconstructed by the second order Kurganov-Tadmor scheme [2]_.

            The default value of $\theta$ in minmod limiter is set to be 1.

        'weno3' - The second order WENO semi-discretization scheme

            .. math ::
                \pdv{u_j}{t}=-\frac{1}{\Delta x}\qty(\hat{f}_{j+1/2}-\hat{f}_{j-1/2})+S(u_j)

            where $\hat{f}_{j+1/2}$ and $\hat{f}_{j-1/2}$ are reconstructed by the weighted essentially non-oscillatory
            scheme [1]_.

    Returns
    =======

    artifact : dict()
        The dict of derived equations and variables.

    References
    ==========

    .. [1] C.-W. Shu, “Essentially non-oscillatory and weighted essentially non- oscillatory schemes for hyperbolic conservation laws,” in Advanced Numerical Approximation of Nonlinear Hyperbolic Equations: Lectures given at the 2nd Session of the Centro Internazionale Matematico Estivo (C.I.M.E.) held in Cetraro, Italy, June 23–28, 1997, A. Quarteroni, Ed. Berlin, Heidelberg: Springer Berlin Heidelberg, 1998, pp. 325–432.

    .. [2] https://doi.org/10.1006/jcph.2000.6459

    .. [3] https://doi.org/10.1109/TSG.2022.3203485

    .. [4] https://doi.org/10.1007/978-3-030-76810-2_2

    .. [5] https://arxiv.org/abs/2410.09464

    """

    if not is_integer(M):
        raise TypeError(f'M is {type(M)} instead of integer')

    if not is_number(dx):
        raise TypeError(f'dx is {type(dx)} instead of number')

    artifact = dict()

    match method:
        case 'euler':
            p0 = AliasVar(p.name, init=p)
            q0 = AliasVar(q.name, init=q)
            artifact[p0.name] = p0
            artifact[q0.name] = q0
            rhs = (p[1:M + 1] - p0[1:M + 1]) / dt + va ** 2 / S * (q[1:M + 1] - q[0:M]) / dx
            artifact[f'euler{pipe_name}_p'] = Eqn(f'euler{pipe_name}_p', rhs)
            rhs = (q[1:M + 1] - q0[1:M + 1]) / dt + S * (p[1:M + 1] - p[0:M]) / dx + lam * va ** 2 * q[1:M + 1] * Abs(
                q[1:M + 1]) / (2 * D * S * p[1:M + 1])
            artifact[f'euler{pipe_name}_q'] = Eqn(f'euler{pipe_name}_q', rhs)

        case 'cdm':
            p0 = AliasVar(p.name, init=p)
            q0 = AliasVar(q.name, init=q)
            artifact[p0.name] = p0
            artifact[q0.name] = q0
            rhs = ((p[1:M + 1] + p[0:M] - p0[1:M + 1] - p0[0:M]) / (2 * dt)
                   + va ** 2 / S * (q[1:M + 1] + q0[1:M + 1] - q[0:M] - q0[0:M]) / (2 * dx))
            artifact[f'cdm{pipe_name}_p'] = Eqn(f'cdm{pipe_name}_p', rhs)
            pba = (p[1:M + 1] + p[0:M] + p0[1:M + 1] + p0[0:M]) / 4
            qba = (q[1:M + 1] + q[0:M] + q0[1:M + 1] + q0[0:M]) / 4
            rhs = ((q[1:M + 1] + q[0:M] - q0[1:M + 1] - q0[0:M]) / (2 * dt)
                   + S * (p[1:M + 1] + p0[1:M + 1] - p[0:M] - p0[0:M]) / (2 * dx)
                   + lam * va ** 2 * qba * Abs(qba) / (2 * D * S * pba))
            artifact[f'cdm{pipe_name}_q'] = Eqn(f'cdm{pipe_name}_q', rhs)

        case 'cha':
            p0 = AliasVar(p.name, init=p)
            q0 = AliasVar(q.name, init=q)
            artifact[p0.name] = p0
            artifact[q0.name] = q0
            p_ = p[1:M + 1]
            q_ = q[1:M + 1]
            p_0 = p0[0:M]
            q_0 = q0[0:M]
            rhs = cha(p_, q_, p_0, q_0, '+', lam, va, S, D, dx)
            artifact[f'cha_{pipe_name}_pos'] = Eqn(f'cha_{pipe_name}_pos', rhs)

            p_ = p[0:M]
            q_ = q[0:M]
            p_0 = p0[1:M + 1]
            q_0 = q0[1:M + 1]
            rhs = cha(p_, q_, p_0, q_0, '-', lam, va, S, D, dx)
            artifact[f'cha_{pipe_name}_neg'] = Eqn(f'cha_{pipe_name}_neg', rhs)
        case 'kt1':
            rhs = mol_tvd1_q_eqn_rhs0([p[0:M - 1], p[1:M], p[2:M + 1]],
                                      [q[0:M - 1], q[1:M], q[2:M + 1]],
                                      S,
                                      va,
                                      lam,
                                      D,
                                      dx)
            artifact['q_' + pipe_name + '_eqn'] = Ode(f'kt1-q{pipe_name}',
                                                      rhs,
                                                      q[1:M])

            rhs = mol_tvd1_p_eqn_rhs0([p[0:M - 1], p[1:M], p[2:M + 1]],
                                      [q[0:M - 1], q[1:M], q[2:M + 1]],
                                      S,
                                      va,
                                      dx)
            artifact['p_' + pipe_name + '_eqn'] = Ode(f'kt1-p{pipe_name}',
                                                      rhs,
                                                      p[1:M])

            artifact['p_' + pipe_name + 'bd1'] = Eqn(p.name + 'bd1',
                                                     S * p[M] + va * q[M] + S * p[M - 2] + va * q[
                                                         M - 2] - 2 * (
                                                             S * p[M - 1] + va * q[M - 1]))
            artifact['q_' + pipe_name + 'bd2'] = Eqn(q.name + 'bd2',
                                                     S * p[2] - va * q[2] + S * p[0] - va * q[0] - 2 * (
                                                             S * p[1] - va * q[1]))

        case 'kt2':
            rhs = mol_tvd1_q_eqn_rhs0([p[0], p[1], p[2]],
                                      [q[0], q[1], q[2]],
                                      S,
                                      va,
                                      lam,
                                      D,
                                      dx)
            artifact['q_' + pipe_name + '_eqn1'] = Ode(f'kt2-q{pipe_name}_1',
                                                       rhs,
                                                       q[1])

            rhs = mol_tvd1_p_eqn_rhs0([p[0], p[1], p[2]],
                                      [q[0], q[1], q[2]],
                                      S,
                                      va,
                                      dx)
            artifact['p_' + pipe_name + '_eqn1'] = Ode(f'kt2-p{pipe_name}_1',
                                                       rhs,
                                                       p[1])

            rhs = mol_tvd1_q_eqn_rhs0([p[M - 2], p[M - 1], p[M]],
                                      [q[M - 2], q[M - 1], q[M]],
                                      S,
                                      va,
                                      lam,
                                      D,
                                      dx)
            artifact['q_' + pipe_name + '_eqn2'] = Ode(f'kt2-q{pipe_name}_2',
                                                       rhs,
                                                       q[M - 1])

            rhs = mol_tvd1_p_eqn_rhs0([p[M - 2], p[M - 1], p[M]],
                                      [q[M - 2], q[M - 1], q[M]],
                                      S,
                                      va,
                                      dx)
            artifact['p_' + pipe_name + '_eqn2'] = Ode(f'kt2-p{pipe_name}_2',
                                                       rhs,
                                                       p[M - 1])

            rhs = mol_tvd2_q_eqn_rhs([p[0:M - 3], p[1:M - 2], p[2:M - 1], p[3:M], p[4:M + 1]],
                                     [q[0:M - 3], q[1:M - 2], q[2:M - 1], q[3:M], q[4:M + 1]],
                                     S,
                                     va,
                                     lam,
                                     D,
                                     dx)
            artifact['q_' + pipe_name + '_eqn3'] = Ode(f'kt2-q{pipe_name}_3',
                                                       rhs,
                                                       q[2:M - 1])

            rhs = mol_tvd2_p_eqn_rhs([p[0:M - 3], p[1:M - 2], p[2:M - 1], p[3:M], p[4:M + 1]],
                                     [q[0:M - 3], q[1:M - 2], q[2:M - 1], q[3:M], q[4:M + 1]],
                                     S,
                                     va,
                                     dx)
            artifact['p_' + pipe_name + '_eqn3'] = Ode(f'kt2-p{pipe_name}_3',
                                                       rhs,
                                                       p[2:M - 1])
            theta = Param('theta', 1)
            artifact['theta'] = theta
            artifact['p_' + pipe_name + 'bd1'] = Eqn(p.name + 'bd1',
                                                     S * p[M] + va * q[M] + S * p[M - 2] + va * q[
                                                         M - 2] - 2 * (
                                                             S * p[M - 1] + va * q[M - 1]))
            artifact['q_' + pipe_name + 'bd2'] = Eqn(q.name + 'bd2',
                                                     S * p[2] - va * q[2] + S * p[0] - va * q[0] - 2 * (
                                                             S * p[1] - va * q[1]))
        case 'weno3':
            rhs = weno_odeq(p[0:M - 3], p[1:M - 2], p[2:M - 1], p[3:M], p[4:M + 1],
                            q[0:M - 3], q[1:M - 2], q[2:M - 1], q[3:M], q[4:M + 1],
                            S, va, lam, D, dx)
            # rhs = mol_weno_q_eqn_rhs(p, q, S, va, lam, D, dx, 2, M - 1)
            # rhs = mol_weno_q_eqn_rhs1([p[0:M - 3], p[1:M - 2], p[2:M - 1], p[3:M], p[4:M + 1]],
            #                           [q[0:M - 3], q[1:M - 2], q[2:M - 1], q[3:M], q[4:M + 1]],
            #                           S,
            #                           va,
            #                           lam,
            #                           D,
            #                           dx)

            artifact['q_' + pipe_name + '_eqn1'] = Ode(f'weno3-q{pipe_name}_1',
                                                       rhs,
                                                       q[2:M - 1])

            rhs = mol_tvd1_q_eqn_rhs1([p[0], p[1], p[2]],
                                      [q[0], q[1], q[2]],
                                      S,
                                      va,
                                      lam,
                                      D,
                                      dx)
            artifact['q_' + pipe_name + '_eqn2'] = Ode(f'weno3-q{pipe_name}_2',
                                                       rhs,
                                                       q[1])
            rhs = mol_tvd1_q_eqn_rhs1([p[M - 2], p[M - 1], p[M]],
                                      [q[M - 2], q[M - 1], q[M]],
                                      S,
                                      va,
                                      lam,
                                      D,
                                      dx)
            artifact['q_' + pipe_name + '_eqn3'] = Ode(f'weno3-q{pipe_name}_3',
                                                       rhs,
                                                       q[M - 1])

            # rhs = mol_weno_p_eqn_rhs(p, q, S, va, dx, 2, M - 1)
            rhs = weno_odep(p[0:M - 3], p[1:M - 2], p[2:M - 1], p[3:M], p[4:M + 1],
                            q[0:M - 3], q[1:M - 2], q[2:M - 1], q[3:M], q[4:M + 1],
                            S, va, lam, D, dx)
            # rhs = mol_weno_p_eqn_rhs1([p[0:M - 3], p[1:M - 2], p[2:M - 1], p[3:M], p[4:M + 1]],
            #                           [q[0:M - 3], q[1:M - 2], q[2:M - 1], q[3:M], q[4:M + 1]],
            #                           S,
            #                           va,
            #                           dx)

            artifact['p_' + pipe_name + '_eqn1'] = Ode(f'weno3-p{pipe_name}_1',
                                                       rhs,
                                                       p[2:M - 1])

            rhs = mol_tvd1_p_eqn_rhs1([p[0], p[1], p[2]],
                                      [q[0], q[1], q[2]],
                                      S,
                                      va,
                                      dx)
            artifact['p_' + pipe_name + '_eqn2'] = Ode(f'weno3-p{pipe_name}_2',
                                                       rhs,
                                                       p[1])
            rhs = mol_tvd1_p_eqn_rhs1([p[M - 2], p[M - 1], p[M]],
                                      [q[M - 2], q[M - 1], q[M]],
                                      S,
                                      va,
                                      dx)
            artifact['p_' + pipe_name + '_eqn3'] = Ode(f'weno3-p{pipe_name}_3',
                                                       rhs,
                                                       p[M - 1])
            artifact['p_' + pipe_name + 'bd1'] = Eqn(p.name + 'bd1',
                                                     S * p[M] + va * q[M] + S * p[M - 2] + va * q[
                                                         M - 2] - 2 * (
                                                             S * p[M - 1] + va * q[M - 1]))
            artifact['q_' + pipe_name + 'bd2'] = Eqn(q.name + 'bd2',
                                                     S * p[2] - va * q[2] + S * p[0] - va * q[0] - 2 * (
                                                             S * p[1] - va * q[1]))
        case _:
            raise NotImplementedError(f'No such method: {method}!')
    return artifact



from Solverz import Eqn, Ode, AliasVar, TimeSeriesParam, Param
from Solverz import iVar, idx, Var, Abs
from Solverz.utilities.type_checker import is_integer, is_number
from ..basic import minmod


def ux(theta, um1, u, up1, dx):
    return minmod(theta * (u - um1) / dx, (up1 - um1) / (2 * dx), theta * (up1 - u) / dx)


def mol_kt2_rhs(T,
                m,
                lam,
                rho,
                Cp,
                S,
                Tamb,
                dx,
                start,
                end,
                order=2):
    """
    Four point stencil
    :param order:
    :param T:
    :param m:
    :param lam:
    :param rho:
    :param Cp:
    :param S:
    :param Tamb:
    :param dx:
    :param start:
    :param end:
    :return:
    """
    id0 = start
    id1 = end
    if id0 >= id1:
        raise ValueError('id0 should be smaller than id1!')

    if end - start == 1:
        Tm2, Tm1, T0, Tp1 = [T[id0 - 2], T[id0 - 1], T[id0], T[id0 + 1]]
    else:
        Tm2, Tm1, T0, Tp1 = [T[id0 - 2:id1 - 2], T[id0 - 1:id1 - 1], T[id0:id1], T[id0 + 1:id1 + 1]]

    theta = Param('theta', 1)
    if order == 2:
        Txj = ux(theta, Tm1, T0, Tp1, dx)
        Txjm1 = ux(theta, Tm2, Tm1, T0, dx)
    elif order == 1:
        Txj = 0
        Txjm1 = 0
    else:
        raise ValueError(f"Order {order} not supported!")

    rhs = m / (S * rho * dx) * (Tm1 + dx / 2 * Txjm1 - T0 - dx / 2 * Txj) - lam / (S * rho * Cp) * (T0 - Tamb)

    return rhs



from SolMuseum.pde.gas.util import *

__all__ = ['rupture_ngs_pipe_cha', 'leakage_ngs_pipe_cha']


def rupture_ngs_pipe_cha(p: Var,
                         q: Var,
                         lam,
                         va,
                         D,
                         S,
                         dx,
                         dt,
                         M,
                         pipe_name: str,
                         idx_leak):
    if not is_integer(M):
        raise TypeError(f'M is {type(M)} instead of integer')

    if not is_number(dx):
        raise TypeError(f'dx is {type(dx)} instead of number')

    artifact = dict()

    qleak1 = Var(f'q_{pipe_name}_leak1', init=q[idx_leak - 1])
    qleak2 = Var(f'q_{pipe_name}_leak2', init=q[idx_leak + 1])
    qleak1_0 = AliasVar(qleak1.name, init=qleak1)
    qleak2_0 = AliasVar(qleak2.name, init=qleak2)

    artifact[qleak1.name] = qleak1
    artifact[qleak2.name] = qleak2
    artifact[qleak1_0.name] = qleak1_0
    artifact[qleak2_0.name] = qleak2_0

    p0 = AliasVar(p.name, init=p)
    q0 = AliasVar(q.name, init=q)
    artifact[p0.name] = p0
    artifact[q0.name] = q0

    rhs = cha(p[1:idx_leak],
              q[1:idx_leak],
              p0[0:idx_leak - 1],
              q0[0:idx_leak - 1],
              '+',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_pos0'] = Eqn(f'cha_{pipe_name}_pos0',
                                            rhs)

    rhs = cha(p[idx_leak],
              qleak1,
              p0[idx_leak - 1],
              q0[idx_leak - 1],
              '+',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_pos1'] = Eqn(f'cha_{pipe_name}_pos1',
                                            rhs)

    rhs = cha(p[idx_leak + 2:M + 1],
              q[idx_leak + 2:M + 1],
              p0[idx_leak + 1:M],
              q0[idx_leak + 1:M],
              '+',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_pos2'] = Eqn(f'cha_{pipe_name}_pos2', rhs)

    rhs = cha(p[idx_leak + 1],
              q[idx_leak + 1],
              p0[idx_leak],
              qleak2_0,
              '+',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_pos3'] = Eqn(f'cha_{pipe_name}_pos3', rhs)

    rhs = cha(p[0:idx_leak - 1],
              q[0:idx_leak - 1],
              p0[1:idx_leak],
              q0[1:idx_leak],
              '-',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_neg0'] = Eqn(f'cha_{pipe_name}_neg0',
                                            rhs)

    rhs = cha(p[idx_leak - 1],
              q[idx_leak - 1],
              p0[idx_leak],
              qleak1_0,
              '-',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_neg1'] = Eqn(f'cha_{pipe_name}_neg1',
                                            rhs)

    rhs = cha(p[idx_leak + 1:M],
              q[idx_leak + 1:M],
              p0[idx_leak + 2:M + 1],
              q0[idx_leak + 2:M + 1],
              '-',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_neg2'] = Eqn(f'cha_{pipe_name}_neg2',
                                            rhs)

    rhs = cha(p[idx_leak],
              qleak2,
              p0[idx_leak + 1],
              q0[idx_leak + 1],
              '-',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_neg3'] = Eqn(f'cha_{pipe_name}_neg3',
                                            rhs)

    pb0 = p.value[idx_leak]
    pa_t = [pb0, pb0, pb0, pb0]
    tseries = [0, 300, 300 + 10, 10 * 3600]
    pa = TimeSeriesParam('pa',
                         v_series=pa_t,
                         time_series=tseries)
    artifact['pa'] = pa

    artifact[f'cha_{pipe_name}_bd1'] = Eqn(f'cha_{pipe_name}_bd1',
                                           p[idx_leak] - pa)
    artifact[f'cha_{pipe_name}_bd2'] = Eqn(f'cha_{pipe_name}_bd2',
                                           qleak1 - qleak2 - q[idx_leak])
    return artifact


def leakage_ngs_pipe_cha(p: Var,
                         q: Var,
                         lam,
                         va,
                         D,
                         S,
                         dx,
                         dt,
                         M,
                         pipe_name: str,
                         idx_leak,
                         d):
    if not is_integer(M):
        raise TypeError(f'M is {type(M)} instead of integer')

    if not is_number(dx):
        raise TypeError(f'dx is {type(dx)} instead of number')

    artifact = dict()

    qleak1 = Var(f'q_{pipe_name}_leak1', init=q[idx_leak - 1])
    qleak2 = Var(f'q_{pipe_name}_leak2', init=q[idx_leak + 1])
    qleak1_0 = AliasVar(qleak1.name, init=qleak1)
    qleak2_0 = AliasVar(qleak2.name, init=qleak2)
    is_sonic = Param('is_sonic', value=1)

    artifact[qleak1.name] = qleak1
    artifact[qleak2.name] = qleak2
    artifact[qleak1_0.name] = qleak1_0
    artifact[qleak2_0.name] = qleak2_0
    artifact[is_sonic.name] = is_sonic

    p0 = AliasVar(p.name, init=p)
    q0 = AliasVar(q.name, init=q)
    artifact[p0.name] = p0
    artifact[q0.name] = q0

    rhs = cha(p[1:idx_leak],
              q[1:idx_leak],
              p0[0:idx_leak - 1],
              q0[0:idx_leak - 1],
              '+',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_pos0'] = Eqn(f'cha_{pipe_name}_pos0',
                                            rhs)

    rhs = cha(p[idx_leak],
              qleak1,
              p0[idx_leak - 1],
              q0[idx_leak - 1],
              '+',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_pos1'] = Eqn(f'cha_{pipe_name}_pos1',
                                            rhs)

    rhs = cha(p[idx_leak + 2:M + 1],
              q[idx_leak + 2:M + 1],
              p0[idx_leak + 1:M],
              q0[idx_leak + 1:M],
              '+',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_pos2'] = Eqn(f'cha_{pipe_name}_pos2', rhs)

    rhs = cha(p[idx_leak + 1],
              q[idx_leak + 1],
              p0[idx_leak],
              qleak2_0,
              '+',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_pos3'] = Eqn(f'cha_{pipe_name}_pos3', rhs)

    rhs = cha(p[0:idx_leak - 1],
              q[0:idx_leak - 1],
              p0[1:idx_leak],
              q0[1:idx_leak],
              '-',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_neg0'] = Eqn(f'cha_{pipe_name}_neg0',
                                            rhs)

    rhs = cha(p[idx_leak - 1],
              q[idx_leak - 1],
              p0[idx_leak],
              qleak1_0,
              '-',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_neg1'] = Eqn(f'cha_{pipe_name}_neg1',
                                            rhs)

    rhs = cha(p[idx_leak + 1:M],
              q[idx_leak + 1:M],
              p0[idx_leak + 2:M + 1],
              q0[idx_leak + 2:M + 1],
              '-',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_neg2'] = Eqn(f'cha_{pipe_name}_neg2',
                                            rhs)

    rhs = cha(p[idx_leak],
              qleak2,
              p0[idx_leak + 1],
              q0[idx_leak + 1],
              '-',
              lam,
              va,
              S,
              D,
              dx)
    artifact[f'cha_{pipe_name}_neg3'] = Eqn(f'cha_{pipe_name}_neg3',
                                            rhs)

    P2 = p[idx_leak]
    leak_rate = TimeSeriesParam('leak_rate',
                                v_series=[0, 0],
                                time_series=[0, 3 * 3600])
    artifact['leak_rate'] = leak_rate

    C0 = 0.61
    T0 = 293
    Mass = 17.1e-3
    Z = 1
    R = 8.314
    Hcr = 1.3
    d_leak = d
    Ah = np.pi * (d_leak / 2) ** 2
    Pa = 101e3
    P2cr = Pa * (2 / (Hcr + 1)) ** (-Hcr / (Hcr - 1))

    qleak = leak_rate * (is_sonic *
                         Ah * P2 *
                         (Mass / (Z * R * T0) * Hcr * (2 / (Hcr + 1)) ** ((Hcr + 1) / (Hcr - 1))) ** (1 / 2)
                         + (1 - is_sonic) *
                         C0 * Ah * P2 *
                         (2 * Mass / (Z * R * T0) * Hcr / (Hcr - 1) * (
                                 (Pa / P2) ** (2 / Hcr) - (Pa / P2) ** ((Hcr + 1) / Hcr))) ** (1 / 2))

    artifact[f'cha_{pipe_name}_bd1'] = Eqn(f'cha_{pipe_name}_bd1',
                                           q[idx_leak] - qleak)
    artifact[f'cha_{pipe_name}_bd2'] = Eqn(f'cha_{pipe_name}_bd2',
                                           qleak1 - qleak2 - q[idx_leak])
    return artifact



from .util import *


def heat_pipe(T: Var,
              m,
              lam,
              rho,
              Cp,
              S,
              Tamb,
              dx,
              dt,
              M,
              pipe_name: str,
              method='kt2'):
    r"""
    This function model and discretize heat pipe equations

        .. math::
            \frac{\partial\tau}{\partial t}+\frac{\dot{m}}{\gamma \rho}\frac{\partial\tau}{\partial x}+
            \frac{\lambda }{\gamma\rho C_\mathrm{p}}(\tau-\tau^\mathrm{amb})=0

    where $\tau$ denotes the two-dimensional temperature distribution, $\gamma$ is the cross sectional ares, $\rho$ is
    the water density, $\dot{m}$ is the water mass flow, $\lambda$ is the friction coefficient, $C_p$ is the thermal
    capacity, and $\tau^\text{amb}$ is the ambient temperature.

    Parameters
    ==========

    T : Var
        The temperature distribution w.r.t. x and t

    m: Param or Var
        The mass flow rate

    lam : Param or Number
        The friction $\lambda$

    rho: Param or Number
        The water density

    Cp: Param or Number
        Thermal capacity of mass flow

    S : Param or Number
        The cross-section area

    Tamb: Param or Number
        The ambient temperature

    dx : Param or Number
        The spatial difference step size

    dt : Param or Number
        The temporal step size. `dt` can be set to 0 if one uses the kt2 scheme.

    M : Param or Number
        The friction $\lambda$

    pipe_name : str
        The name of the pipe

    method : str

        'iu' - The method of implicit upwind [1]_

            .. math ::
                \left\{
                \begin{aligned}
                    &\frac{\partial \tau}{\partial t}=\frac{\tau_{k+1}^{n+1}-\tau_{k+1}^{n}}{\Delta t}\\
                    &\frac{\partial \tau}{\partial x}=\frac{\tau_{k+1}^{n+1}-\tau_k^{n+1}}{\Delta x}\\
                    &\tau=\tau_{k+1}^{n+1}
                \end{aligned}
                \right.

        'yao' - The Yao's scheme, i.e. the second order explicit scheme [2]_

            .. math ::
                \left\{
                \begin{aligned}
                    &\frac{\partial \tau}{\partial t}=\frac{\tau_k^{n+1}-\tau_k^{n}+\tau_{k+1}^{n+1}-\tau_{k+1}^{n}}{2\Delta t}\\
                    &\frac{\partial \tau}{\partial x}=\frac{\tau_{k+1}^{n+1}-\tau_k^{n+1}+\tau_{k+1}^{n}-\tau_{k}^{n}}{2\Delta x}\\
                     &\tau=\frac{\tau_{k+1}^{n+1}+\tau_k^{n+1}+\tau_{k+1}^{n}+\tau_{k}^{n}}{4}
                \end{aligned}
                \right.

        'kt2' - Default, the second order kurganov-tadmor scheme [3]_

            .. math ::
                \pdv{u_j}{t}=-\frac{1}{\Delta x}\qty(\hat{f}_{j+1/2}-\hat{f}_{j-1/2})+S(u_j)

            where $\hat{f}_{j+1/2}$ and $\hat{f}_{j-1/2}$ are reconstructed by the second order kurganov-tadmor scheme.

    Returns
    =======

    artifact : dict()
        The dict of derived equations and variables.

    References
    ==========

    .. [1] https://doi.org/10.1016/j.apenergy.2017.08.061

    .. [2] https://doi.org/10.1109/TSTE.2020.2988682

    .. [3] https://doi.org/10.1006/jcph.2000.6459

    """

    if not is_integer(M):
        raise TypeError(f'M is {type(M)} instead of integer')

    if not is_number(dx):
        raise TypeError(f'dx is {type(dx)} instead of number')

    artifact = dict()

    match method:
        case 'kt2':
            rhs = mol_kt2_rhs(T,
                              m,
                              lam,
                              rho,
                              Cp,
                              S,
                              Tamb,
                              dx,
                              1,
                              2,
                              order=1)
            artifact['T' + pipe_name + '_eqn1'] = Ode(f'heat_pipe_kt2_T{pipe_name}_1',
                                                      rhs,
                                                      T[1])
            rhs = mol_kt2_rhs(T,
                              m,
                              lam,
                              rho,
                              Cp,
                              S,
                              Tamb,
                              dx,
                              2,
                              M,
                              order=2)
            artifact['T' + pipe_name + '_eqn2'] = Ode(f'heat_pipe_kt2_T{pipe_name}_2',
                                                      rhs,
                                                      T[2:M])
            rhs = mol_kt2_rhs(T,
                              m,
                              lam,
                              rho,
                              Cp,
                              S,
                              Tamb,
                              dx,
                              M,
                              M+1,
                              order=1)
            artifact['T' + pipe_name + '_eqn3'] = Ode(f'heat_pipe_kt2_T{pipe_name}_3',
                                                      rhs,
                                                      T[M])
            artifact['theta'] = Param('theta', 1)
        case 'iu':
            if is_number(dt):
                if dt <= 0:
                    raise ValueError(f'dt must be positive, got {dt}')
            T0 = AliasVar(T.name, init=T)
            artifact[T0.name] = T0
            dTdt = (T[1:M+1]-T0[1:M+1])/dt
            dTdx = (T[1:M+1]-T[0:M])/dx
            rhs = dTdt + m/S/rho * dTdx + lam/S/rho/Cp*(T[1:M+1]-Tamb)
            artifact[f'iu_pipe_{pipe_name}'] = Eqn(f'iu_pipe_{pipe_name}', rhs)
        case 'yao':
            if is_number(dt):
                if dt <= 0:
                    raise ValueError(f'dt must be positive, got {dt}')
            T0 = AliasVar(T.name, init=T)
            artifact[T0.name] = T0
            dTdt = (T[1:M+1]-T0[1:M+1]+T[0:M]-T0[0:M])/(2*dt)
            dTdx = (T[1:M+1]+T0[1:M+1]-T[0:M]-T0[0:M])/(2*dx)
            Tavg = (T[1:M+1]+T0[1:M+1]+T[0:M]+T0[0:M])/4
            rhs = dTdt + m/S/rho*dTdx + lam/S/rho/Cp*(Tavg-Tamb)
            artifact[f'yao_pipe_{pipe_name}'] = Eqn(f'yao_pipe_{pipe_name}', rhs)
        case _:
            raise NotImplementedError(f'No such method: {method}!')
    return artifact



from SolMuseum.pde.gas.util import *

__all__ = ['rupture_ngs_pipe_kt2', 'leakage_ngs_pipe_kt2']


def rupture_ngs_pipe_kt2(p: Var,
                         q: Var,
                         lam,
                         va,
                         D,
                         S,
                         dx,
                         M,
                         pipe_name: str,
                         idx_leak):
    if not is_integer(M):
        raise TypeError(f'M is {type(M)} instead of integer')

    if not is_number(dx):
        raise TypeError(f'dx is {type(dx)} instead of number')

    artifact = dict()

    qleak1 = Var(f'q_{pipe_name}_leak1', init=q[idx_leak - 1])
    qleak2 = Var(f'q_{pipe_name}_leak2', init=q[idx_leak + 1])
    CUT_DOWNSTREAM = Param('CUT_DOWNSTREAM', 0)

    artifact[qleak1.name] = qleak1
    artifact[qleak2.name] = qleak2
    artifact[CUT_DOWNSTREAM.name] = CUT_DOWNSTREAM

    # index = 0
    artifact['q' + pipe_name + 'bd2'] = Eqn('q' + pipe_name + 'bd2',
                                            S * p[2] - va * q[2] + S * p[0] - va * q[0] - 2 * (
                                                    S * p[1] - va * q[1]))
    # index = 1
    rhs = mol_tvd1_q_eqn_rhs1([p[0], p[1], p[2]],
                              [q[0], q[1], q[2]],
                              S,
                              va,
                              lam,
                              D,
                              dx)
    artifact['q' + pipe_name + '_eqn2'] = Ode(f'kt2-q_{pipe_name}_2',
                                              rhs,
                                              q[1])
    rhs = mol_tvd1_p_eqn_rhs1([p[0], p[1], p[2]],
                              [q[0], q[1], q[2]],
                              S,
                              va,
                              dx)
    artifact['p' + pipe_name + '_eqn2'] = Ode(f'kt2-p_{pipe_name}_2',
                                              rhs,
                                              p[1])
    # 2<= index <= idx_leak-3
    rhs = mol_tvd2_q_eqn_rhs(
        [p[0:idx_leak - 4], p[1:idx_leak - 3], p[2:idx_leak - 2], p[3:idx_leak - 1], p[4:idx_leak]],
        [q[0:idx_leak - 4], q[1:idx_leak - 3], q[2:idx_leak - 2], q[3:idx_leak - 1], q[4:idx_leak]],
        S,
        va,
        lam,
        D,
        dx)
    artifact['q' + pipe_name + '_leak1'] = Ode(f'kt2-q_{pipe_name}__leak_1',
                                               rhs,
                                               q[2:idx_leak - 2])
    rhs = mol_tvd2_p_eqn_rhs(
        [p[0:idx_leak - 4], p[1:idx_leak - 3], p[2:idx_leak - 2], p[3:idx_leak - 1], p[4:idx_leak]],
        [q[0:idx_leak - 4], q[1:idx_leak - 3], q[2:idx_leak - 2], q[3:idx_leak - 1], q[4:idx_leak]],
        S,
        va,
        dx)
    artifact['p' + pipe_name + '_leak1'] = Ode(f'kt2-p_{pipe_name}__leak_1',
                                               rhs,
                                               p[2:idx_leak - 2])

    # index = idx_leak-2
    rhs = mol_tvd2_q_eqn_rhs([p[idx_leak - 4], p[idx_leak - 3], p[idx_leak - 2], p[idx_leak - 1], p[idx_leak]],
                             [q[idx_leak - 4], q[idx_leak - 3], q[idx_leak - 2], q[idx_leak - 1], qleak1],
                             S,
                             va,
                             lam,
                             D,
                             dx)
    artifact['q' + pipe_name + '_leak6'] = Ode(f'kt2-q_{pipe_name}__leak_6',
                                               rhs,
                                               q[idx_leak - 2])
    rhs = mol_tvd2_p_eqn_rhs([p[idx_leak - 4], p[idx_leak - 3], p[idx_leak - 2], p[idx_leak - 1], p[idx_leak]],
                             [q[idx_leak - 4], q[idx_leak - 3], q[idx_leak - 2], q[idx_leak - 1], qleak1],
                             S,
                             va,
                             dx)
    artifact['p' + pipe_name + '_leak6'] = Ode(f'kt2-p_{pipe_name}__leak_6',
                                               rhs,
                                               p[idx_leak - 2])

    # index = idx_leak-1
    rhs = mol_tvd1_q_eqn_rhs1([p[idx_leak - 2], p[idx_leak - 1], p[idx_leak]],
                              [q[idx_leak - 2], q[idx_leak - 1], qleak1],
                              S,
                              va,
                              lam,
                              D,
                              dx)
    artifact['q' + pipe_name + '_leak2'] = Ode(f'kt2-q_{pipe_name}__leak_2',
                                               rhs,
                                               q[idx_leak - 1])
    rhs = mol_tvd1_p_eqn_rhs1([p[idx_leak - 2], p[idx_leak - 1], p[idx_leak]],
                              [q[idx_leak - 2], q[idx_leak - 1], qleak1],
                              S,
                              va,
                              dx)
    artifact['p' + pipe_name + '_leak2'] = Ode(f'kt2-p_{pipe_name}__leak_2',
                                               rhs,
                                               p[idx_leak - 1])
    # index = idx_leak
    eqn = S * p[idx_leak] + va * qleak1 + S * p[idx_leak - 2] + va * q[idx_leak - 2] - 2 * (
            S * p[idx_leak - 1] + va * q[idx_leak - 1])
    artifact['q' + pipe_name + '_leak3_1'] = Eqn(f'kt2-q_{pipe_name}__leak_3_1',
                                                 eqn)
    eqn = (S * p[idx_leak + 2] - va * q[idx_leak + 2] + S * p[idx_leak] - va * qleak2
           - 2 * (S * p[idx_leak + 1] - va * q[idx_leak + 1]))
    artifact['q' + pipe_name + '_leak3_2'] = Eqn(f'kt2-q_{pipe_name}__leak_3_2',
                                                 eqn)

    # index = idx_leak+1
    rhs = mol_tvd1_q_eqn_rhs1([p[idx_leak], p[idx_leak + 1], p[idx_leak + 2]],
                              [qleak2, q[idx_leak + 1], q[idx_leak + 2]],
                              S,
                              va,
                              lam,
                              D,
                              dx)
    artifact['q' + pipe_name + '_leak4'] = Ode(f'kt2-q_{pipe_name}__leak_4',
                                               rhs,
                                               q[idx_leak + 1])
    rhs = mol_tvd1_p_eqn_rhs1([p[idx_leak], p[idx_leak + 1], p[idx_leak + 2]],
                              [qleak2, q[idx_leak + 1], q[idx_leak + 2]],
                              S,
                              va,
                              dx)
    artifact['p' + pipe_name + '_leak4'] = Ode(f'kt2-p_{pipe_name}__leak_4',
                                               rhs,
                                               p[idx_leak + 1])

    #  index = idx_leak + 2
    rhs = mol_tvd2_q_eqn_rhs([p[idx_leak], p[idx_leak + 1], p[idx_leak + 2], p[idx_leak + 3], p[idx_leak + 4]],
                             [qleak2, q[idx_leak + 1], q[idx_leak + 2], q[idx_leak + 3], q[idx_leak + 4]],
                             S,
                             va,
                             lam,
                             D,
                             dx)
    artifact['q' + pipe_name + '_leak7'] = Ode(f'kt2-q_{pipe_name}__leak_7',
                                               rhs,
                                               q[idx_leak + 2])
    rhs = mol_tvd2_p_eqn_rhs([p[idx_leak], p[idx_leak + 1], p[idx_leak + 2], p[idx_leak + 3], p[idx_leak + 4]],
                             [qleak2, q[idx_leak + 1], q[idx_leak + 2], q[idx_leak + 3], q[idx_leak + 4]],
                             S,
                             va,
                             dx)
    artifact['p' + pipe_name + '_leak7'] = Ode(f'kt2-p_{pipe_name}__leak_7',
                                               rhs,
                                               p[idx_leak + 2])

    #  M-2>=index > idx_leak + 2
    rhs = mol_tvd2_q_eqn_rhs(
        [p[idx_leak + 1:M - 3], p[idx_leak + 2:M - 2], p[idx_leak + 3:M - 1], p[idx_leak + 4:M], p[idx_leak + 5:M + 1]],
        [q[idx_leak + 1:M - 3], q[idx_leak + 2:M - 2], q[idx_leak + 3:M - 1], q[idx_leak + 4:M], q[idx_leak + 5:M + 1]],
        S,
        va,
        lam,
        D,
        dx)
    artifact['q' + pipe_name + '_leak5'] = Ode(f'kt2-q_{pipe_name}__leak_5',
                                               rhs,
                                               q[idx_leak + 3: M - 1])
    rhs = mol_tvd2_p_eqn_rhs(
        [p[idx_leak + 1:M - 3], p[idx_leak + 2:M - 2], p[idx_leak + 3:M - 1], p[idx_leak + 4:M], p[idx_leak + 5:M + 1]],
        [q[idx_leak + 1:M - 3], q[idx_leak + 2:M - 2], q[idx_leak + 3:M - 1], q[idx_leak + 4:M], q[idx_leak + 5:M + 1]],
        S,
        va,
        dx)
    artifact['p' + pipe_name + '_leak5'] = Ode(f'kt2-p_{pipe_name}__leak_5',
                                               rhs,
                                               p[idx_leak + 3:M - 1])
    # index = M-1
    rhs = mol_tvd1_q_eqn_rhs1([p[M - 2], p[M - 1], p[M]],
                              [q[M - 2], q[M - 1], q[M]],
                              S,
                              va,
                              lam,
                              D,
                              dx)
    artifact['q' + pipe_name + '_eqn3'] = Ode(f'kt2-q_{pipe_name}_3',
                                              rhs,
                                              q[M - 1])
    rhs = mol_tvd1_p_eqn_rhs1([p[M - 2], p[M - 1], p[M]],
                              [q[M - 2], q[M - 1], q[M]],
                              S,
                              va,
                              dx)
    artifact['p' + pipe_name + '_eqn3'] = Ode(f'kt2-p_{pipe_name}_3',
                                              rhs,
                                              p[M - 1])

    # index = M
    artifact['p' + pipe_name + 'bd1'] = Eqn('p' + pipe_name + 'bd1',
                                            S * p[M] + va * q[M] + S * p[M - 2] + va * q[
                                                M - 2] - 2 * (S * p[M - 1] + va * q[M - 1]))

    pb0 = p.value[idx_leak]
    pa_t = [pb0, pb0, pb0, pb0]
    tseries = [0, 300, 300 + 10, 10 * 3600]
    pa = TimeSeriesParam('pa',
                         v_series=pa_t,
                         time_series=tseries)
    artifact['pa'] = pa

    artifact[f'kt2_{pipe_name}_bd1'] = Eqn(f'pipe_{pipe_name}_bd1',
                                            p[idx_leak] - pa)
    artifact[f'kt2_{pipe_name}_bd2'] = Eqn(f'pipe_{pipe_name}_bd2',
                                            qleak1 - qleak2 - q[idx_leak])
    return artifact


def leakage_ngs_pipe_kt2(p: Var,
                         q: Var,
                         lam,
                         va,
                         D,
                         S,
                         dx,
                         M,
                         pipe_name: str,
                         idx_leak,
                         d):
    if not is_integer(M):
        raise TypeError(f'M is {type(M)} instead of integer')

    if not is_number(dx):
        raise TypeError(f'dx is {type(dx)} instead of number')

    artifact = dict()

    qleak1 = Var(f'q_{pipe_name}_leak1', init=q[idx_leak - 1])
    qleak2 = Var(f'q_{pipe_name}_leak2', init=q[idx_leak + 1])

    artifact[qleak1.name] = qleak1
    artifact[qleak2.name] = qleak2

    is_sonic = Param('is_sonic', value=1)
    artifact[is_sonic.name] = is_sonic

    # index = 0
    artifact['q' + pipe_name + 'bd2'] = Eqn('q' + pipe_name + 'bd2',
                                            S * p[2] - va * q[2] + S * p[0] - va * q[0] - 2 * (
                                                    S * p[1] - va * q[1]))
    # index = 1
    rhs = mol_tvd1_q_eqn_rhs1([p[0], p[1], p[2]],
                              [q[0], q[1], q[2]],
                              S,
                              va,
                              lam,
                              D,
                              dx)
    artifact['q' + pipe_name + '_eqn2'] = Ode(f'kt2-q_{pipe_name}_2',
                                              rhs,
                                              q[1])
    rhs = mol_tvd1_p_eqn_rhs1([p[0], p[1], p[2]],
                              [q[0], q[1], q[2]],
                              S,
                              va,
                              dx)
    artifact['p' + pipe_name + '_eqn2'] = Ode(f'kt2-p_{pipe_name}_2',
                                              rhs,
                                              p[1])
    # 2<= index <= idx_leak-3
    rhs = mol_tvd2_q_eqn_rhs(
        [p[0:idx_leak - 4], p[1:idx_leak - 3], p[2:idx_leak - 2], p[3:idx_leak - 1], p[4:idx_leak]],
        [q[0:idx_leak - 4], q[1:idx_leak - 3], q[2:idx_leak - 2], q[3:idx_leak - 1], q[4:idx_leak]],
        S,
        va,
        lam,
        D,
        dx)
    artifact['q' + pipe_name + '_leak1'] = Ode(f'kt2-q_{pipe_name}__leak_1',
                                               rhs,
                                               q[2:idx_leak - 2])
    rhs = mol_tvd2_p_eqn_rhs(
        [p[0:idx_leak - 4], p[1:idx_leak - 3], p[2:idx_leak - 2], p[3:idx_leak - 1], p[4:idx_leak]],
        [q[0:idx_leak - 4], q[1:idx_leak - 3], q[2:idx_leak - 2], q[3:idx_leak - 1], q[4:idx_leak]],
        S,
        va,
        dx)
    artifact['p' + pipe_name + '_leak1'] = Ode(f'kt2-p_{pipe_name}__leak_1',
                                               rhs,
                                               p[2:idx_leak - 2])

    # index = idx_leak-2
    rhs = mol_tvd2_q_eqn_rhs([p[idx_leak - 4], p[idx_leak - 3], p[idx_leak - 2], p[idx_leak - 1], p[idx_leak]],
                             [q[idx_leak - 4], q[idx_leak - 3], q[idx_leak - 2], q[idx_leak - 1], qleak1],
                             S,
                             va,
                             lam,
                             D,
                             dx)
    artifact['q' + pipe_name + '_leak6'] = Ode(f'kt2-q_{pipe_name}__leak_6',
                                               rhs,
                                               q[idx_leak - 2])
    rhs = mol_tvd2_p_eqn_rhs([p[idx_leak - 4], p[idx_leak - 3], p[idx_leak - 2], p[idx_leak - 1], p[idx_leak]],
                             [q[idx_leak - 4], q[idx_leak - 3], q[idx_leak - 2], q[idx_leak - 1], qleak1],
                             S,
                             va,
                             dx)
    artifact['p' + pipe_name + '_leak6'] = Ode(f'kt2-p_{pipe_name}__leak_6',
                                               rhs,
                                               p[idx_leak - 2])

    # index = idx_leak-1
    rhs = mol_tvd1_q_eqn_rhs1([p[idx_leak - 2], p[idx_leak - 1], p[idx_leak]],
                              [q[idx_leak - 2], q[idx_leak - 1], qleak1],
                              S,
                              va,
                              lam,
                              D,
                              dx)
    artifact['q' + pipe_name + '_leak2'] = Ode(f'kt2-q_{pipe_name}__leak_2',
                                               rhs,
                                               q[idx_leak - 1])
    rhs = mol_tvd1_p_eqn_rhs1([p[idx_leak - 2], p[idx_leak - 1], p[idx_leak]],
                              [q[idx_leak - 2], q[idx_leak - 1], qleak1],
                              S,
                              va,
                              dx)
    artifact['p' + pipe_name + '_leak2'] = Ode(f'kt2-p_{pipe_name}__leak_2',
                                               rhs,
                                               p[idx_leak - 1])
    # index = idx_leak
    eqn = S * p[idx_leak] + va * qleak1 + S * p[idx_leak - 2] + va * q[idx_leak - 2] - 2 * (
            S * p[idx_leak - 1] + va * q[idx_leak - 1])
    artifact['q' + pipe_name + '_leak3_1'] = Eqn(f'kt2-q_{pipe_name}__leak_3_1',
                                                 eqn)
    eqn = (S * p[idx_leak + 2] - va * q[idx_leak + 2] + S * p[idx_leak] - va * qleak2
           - 2 * (S * p[idx_leak + 1] - va * q[idx_leak + 1]))
    artifact['q' + pipe_name + '_leak3_2'] = Eqn(f'kt2-q_{pipe_name}__leak_3_2',
                                                 eqn)

    # index = idx_leak+1
    rhs = mol_tvd1_q_eqn_rhs1([p[idx_leak], p[idx_leak + 1], p[idx_leak + 2]],
                              [qleak2, q[idx_leak + 1], q[idx_leak + 2]],
                              S,
                              va,
                              lam,
                              D,
                              dx)
    artifact['q' + pipe_name + '_leak4'] = Ode(f'kt2-q_{pipe_name}__leak_4',
                                               rhs,
                                               q[idx_leak + 1])
    rhs = mol_tvd1_p_eqn_rhs1([p[idx_leak], p[idx_leak + 1], p[idx_leak + 2]],
                              [qleak2, q[idx_leak + 1], q[idx_leak + 2]],
                              S,
                              va,
                              dx)
    artifact['p' + pipe_name + '_leak4'] = Ode(f'kt2-p_{pipe_name}__leak_4',
                                               rhs,
                                               p[idx_leak + 1])

    #  index = idx_leak + 2
    rhs = mol_tvd2_q_eqn_rhs([p[idx_leak], p[idx_leak + 1], p[idx_leak + 2], p[idx_leak + 3], p[idx_leak + 4]],
                             [qleak2, q[idx_leak + 1], q[idx_leak + 2], q[idx_leak + 3], q[idx_leak + 4]],
                             S,
                             va,
                             lam,
                             D,
                             dx)
    artifact['q' + pipe_name + '_leak7'] = Ode(f'kt2-q_{pipe_name}__leak_7',
                                               rhs,
                                               q[idx_leak + 2])
    rhs = mol_tvd2_p_eqn_rhs([p[idx_leak], p[idx_leak + 1], p[idx_leak + 2], p[idx_leak + 3], p[idx_leak + 4]],
                             [qleak2, q[idx_leak + 1], q[idx_leak + 2], q[idx_leak + 3], q[idx_leak + 4]],
                             S,
                             va,
                             dx)
    artifact['p' + pipe_name + '_leak7'] = Ode(f'kt2-p_{pipe_name}__leak_7',
                                               rhs,
                                               p[idx_leak + 2])

    #  M-2>=index > idx_leak + 2
    rhs = mol_tvd2_q_eqn_rhs(
        [p[idx_leak + 1:M - 3], p[idx_leak + 2:M - 2], p[idx_leak + 3:M - 1], p[idx_leak + 4:M], p[idx_leak + 5:M + 1]],
        [q[idx_leak + 1:M - 3], q[idx_leak + 2:M - 2], q[idx_leak + 3:M - 1], q[idx_leak + 4:M], q[idx_leak + 5:M + 1]],
        S,
        va,
        lam,
        D,
        dx)
    artifact['q' + pipe_name + '_leak5'] = Ode(f'kt2-q_{pipe_name}__leak_5',
                                               rhs,
                                               q[idx_leak + 3: M - 1])
    rhs = mol_tvd2_p_eqn_rhs(
        [p[idx_leak + 1:M - 3], p[idx_leak + 2:M - 2], p[idx_leak + 3:M - 1], p[idx_leak + 4:M], p[idx_leak + 5:M + 1]],
        [q[idx_leak + 1:M - 3], q[idx_leak + 2:M - 2], q[idx_leak + 3:M - 1], q[idx_leak + 4:M], q[idx_leak + 5:M + 1]],
        S,
        va,
        dx)
    artifact['p' + pipe_name + '_leak5'] = Ode(f'kt2-p_{pipe_name}__leak_5',
                                               rhs,
                                               p[idx_leak + 3:M - 1])
    # index = M-1
    rhs = mol_tvd1_q_eqn_rhs1([p[M - 2], p[M - 1], p[M]],
                              [q[M - 2], q[M - 1], q[M]],
                              S,
                              va,
                              lam,
                              D,
                              dx)
    artifact['q' + pipe_name + '_eqn3'] = Ode(f'kt2-q_{pipe_name}_3',
                                              rhs,
                                              q[M - 1])
    rhs = mol_tvd1_p_eqn_rhs1([p[M - 2], p[M - 1], p[M]],
                              [q[M - 2], q[M - 1], q[M]],
                              S,
                              va,
                              dx)
    artifact['p' + pipe_name + '_eqn3'] = Ode(f'kt2-p_{pipe_name}_3',
                                              rhs,
                                              p[M - 1])

    # index = M
    artifact['p' + pipe_name + 'bd1'] = Eqn('p' + pipe_name + 'bd1',
                                            S * p[M] + va * q[M] + S * p[M - 2] + va * q[
                                                M - 2] - 2 * (S * p[M - 1] + va * q[M - 1]))

    P2 = p[idx_leak]
    leak_rate = TimeSeriesParam('leak_rate',
                                v_series=[0, 0],
                                time_series=[0, 3 * 3600])
    artifact['leak_rate'] = leak_rate

    C0 = 0.61
    T0 = 293
    Mass = 17.1e-3
    Z = 1
    R = 8.314
    Hcr = 1.3
    d_leak = d
    Ah = np.pi * (d_leak / 2) ** 2
    Pa = 101e3
    P2cr = Pa * (2 / (Hcr + 1)) ** (-Hcr / (Hcr - 1))

    qjleak = leak_rate * (is_sonic *
                          Ah * P2 *
                          (Mass / (Z * R * T0) * Hcr * (2 / (Hcr + 1)) ** ((Hcr + 1) / (Hcr - 1))) ** (1 / 2)
                          + (1 - is_sonic) *
                          C0 * Ah * P2 *
                          (2 * Mass / (Z * R * T0) * Hcr / (Hcr - 1) * (
                                  (Pa / P2) ** (2 / Hcr) - (Pa / P2) ** ((Hcr + 1) / Hcr))) ** (1 / 2))

    artifact[f'kt2_{pipe_name}_bd1'] = Eqn(f'kt2_{pipe_name}_bd1',
                                           q[idx_leak] - qjleak)
    artifact[f'kt2_{pipe_name}_bd2'] = Eqn(f'kt2_{pipe_name}_bd2',
                                           qleak1 - qleak2 - q[idx_leak])
    return artifact



from SolMuseum.pde.gas.util import *
from SolMuseum.pde.basic import SolPde

variables = ['pm2', 'pm1', 'p0', 'pp1', 'pp2', 'qm2', 'qm1', 'q0', 'qp1', 'qp2']
dweno_odepd = [type(f'dweno_odepd{varname}', (SolPde,), {}) for varname in variables]
dweno_odeqd = [type(f'dweno_odeqd{varname}', (SolPde,), {}) for varname in variables]

for i, varname in enumerate(variables):
    setattr(dweno_odepd[i], 'arglength', 15)
    setattr(dweno_odeqd[i], 'arglength', 15)


class weno_odep(SolPde):
    """
    pm2, pm1, p0, pp1, pp2, qm2, qm1, q0, qp1, qp2, S, va, lam, D, dx
    """

    arglength = 15

    def fdiff(self, argindex=1):
        if argindex in range(1, len(variables) + 1):
            return dweno_odepd[argindex - 1](*self.args)
        else:
            return Integer(0)


class weno_odeq(SolPde):
    """
    pm2, pm1, p0, pp1, pp2, qm2, qm1, q0, qp1, qp2, S, va, lam, D, dx
    """

    arglength = 15

    def fdiff(self, argindex=1):
        if argindex in range(1, len(variables) + 1):
            return dweno_odeqd[argindex - 1](*self.args)
        else:
            return Integer(0)



from SolMuseum.pde.gas.util import *
from .euler_broken import rupture_ngs_pipe_Euler, leakage_ngs_pipe_Euler
from .cdm_broken import rupture_ngs_pipe_cdm, leakage_ngs_pipe_cdm
from .kt1_broken import rupture_ngs_pipe_kt1, leakage_ngs_pipe_kt1
from .kt2_broken import rupture_ngs_pipe_kt2, leakage_ngs_pipe_kt2
from .cha_broken import rupture_ngs_pipe_cha, leakage_ngs_pipe_cha
from .weno3_broken import rupture_ngs_pipe_weno3, leakage_ngs_pipe_weno3


def rupture_pipe(p: Var,
                 q: Var,
                 lam,
                 va,
                 D,
                 S,
                 dx,
                 dt,
                 M,
                 pipe_name: str,
                 idx_leak,
                 method='weno3'):
    r"""
    The rupture fault model of ngs pipes, with the fault boundary condition

        .. math::
            \begin{aligned}
            &p_i = p_a\\
            &q_\text{i} = q_\text{upstream} - q_\text{downstream}
            \end{aligned}

    where $p_i$ is the pressure of the rupture position, $p_a$ is the atmospheric pressure, $q_i$ is the mass flow of
    the rupture position, $q_\text{upstream}$ is the upstream leakage mass flow and $q_\text{downstream}$ is the downstream
    leakage mass flow.
    The fault model is based on the normal ngs pipe defined above. The details can be found in [1]_.

    Parameters
    ==========

    idx_leak: Number

        The index of rupture location.

    method : str

        'euler' - Default, the euler scheme

        'cdm' - The central difference scheme

        'cha' - The method of characteristics

        'kt1' - The first order Kurganov-Tadmor semi-discretization scheme

        'kt2' - The second order Kurganov-Tadmor semi-discretization scheme

        'weno3' - The third order WENO semi-discretization scheme

    Returns
    =======

    artifact : dict()
        The dict of derived equations and variables.

    References
    ==========

    .. [1] https://arxiv.org/abs/2410.09464

    """

    if not is_integer(M):
        raise TypeError(f'M is {type(M)} instead of integer')

    if not is_number(dx):
        raise TypeError(f'dx is {type(dx)} instead of number')

    if not is_number(idx_leak):
        raise TypeError(f'idx_leak is {type(idx_leak)} instead of number')

    match method:
        case 'euler':
            return rupture_ngs_pipe_Euler(p,
                                          q,
                                          lam,
                                          va,
                                          D,
                                          S,
                                          dx,
                                          dt,
                                          M,
                                          pipe_name,
                                          idx_leak)

        case 'cdm':
            return rupture_ngs_pipe_cdm(p,
                                        q,
                                        lam,
                                        va,
                                        D,
                                        S,
                                        dx,
                                        dt,
                                        M,
                                        pipe_name,
                                        idx_leak)

        case 'cha':
            return rupture_ngs_pipe_cha(p,
                                        q,
                                        lam,
                                        va,
                                        D,
                                        S,
                                        dx,
                                        dt,
                                        M,
                                        pipe_name,
                                        idx_leak)
        case 'kt1':
            return rupture_ngs_pipe_kt1(p,
                                        q,
                                        lam,
                                        va,
                                        D,
                                        S,
                                        dx,
                                        M,
                                        pipe_name,
                                        idx_leak)

        case 'kt2':
            return rupture_ngs_pipe_kt2(p,
                                        q,
                                        lam,
                                        va,
                                        D,
                                        S,
                                        dx,
                                        M,
                                        pipe_name,
                                        idx_leak)
        case 'weno3':
            return rupture_ngs_pipe_weno3(p,
                                          q,
                                          lam,
                                          va,
                                          D,
                                          S,
                                          dx,
                                          M,
                                          pipe_name,
                                          idx_leak)
        case _:
            raise NotImplementedError(f'No such method: {method}!')


def leakage_pipe(p: Var,
                 q: Var,
                 lam,
                 va,
                 D,
                 S,
                 dx,
                 dt,
                 M,
                 pipe_name: str,
                 idx_leak,
                 d,
                 method='weno3'):
    r"""
    The leakage fault model of ngs pipes, with the fault boundary condition

        .. math::
            q_\text{leak}=
            \begin{cases}
                0.61S_\text{h}p_j\sqrt{\frac{2M}{ZRT}\frac{k}{k-1}\left(\left(\frac{p_\text{a}}{p_j}\right)^{\frac{2}{k}}-\left(\frac{p_\text{a}}{p_j}\right)^{\frac{k+1}{k}}\right)}&p_j\leq {p_\text{sw}}\\
                S_\text{h}p_j\sqrt{\frac{M}{ZRT}k\left(\frac{2}{k+1}\right)^{\frac{k+1}{k-1}}} & p_j > {p_\text{sw}}\\
            \end{cases}

    where $k$ is the adiabatic coefficient, $M$ is the molar mass of gas, $T$ is 273.15K, $Z$ is the gas
    compressibility factor, $R$ is the ideal gas constant, $S_\text{h}$ is the area of the leakage hole;

        .. math::
            S_\text{h} = \pi\qty(\frac{d}{2})^2,

    $d$ is the diameter of the leakage hole; the switching pressure

        .. math::
            p_\text{sw}=\left(\frac{2}{k+1}\right)^{-k/(k-1)}\cdot p_{a}.

    The fault model is based on the normal ngs pipe defined above.

    Parameters
    ==========

    idx_leak: Number

        The index of rupture location.

    d: Number or Param

        The diameter of the leakage hole.

    method : str

        'euler' - Default, the euler scheme

        'cdm' - The central difference scheme

        'cha' - The method of characteristics

        'kt1' - The first order Kurganov-Tadmor semi-discretization scheme

        'kt2' - The second order Kurganov-Tadmor semi-discretization scheme

        'weno3' - The third order WENO semi-discretization scheme

    Returns
    =======

    artifact : dict()
        The dict of derived equations and variables.

    """

    if not is_integer(M):
        raise TypeError(f'M is {type(M)} instead of integer')

    if not is_number(dx):
        raise TypeError(f'dx is {type(dx)} instead of number')

    if not is_number(idx_leak):
        raise TypeError(f'idx_leak is {type(idx_leak)} instead of number')

    match method:
        case 'euler':
            return leakage_ngs_pipe_Euler(p,
                                          q,
                                          lam,
                                          va,
                                          D,
                                          S,
                                          dx,
                                          dt,
                                          M,
                                          pipe_name,
                                          idx_leak,
                                          d)

        case 'cdm':
            return leakage_ngs_pipe_cdm(p,
                                        q,
                                        lam,
                                        va,
                                        D,
                                        S,
                                        dx,
                                        dt,
                                        M,
                                        pipe_name,
                                        idx_leak,
                                        d)

        case 'cha':
            return leakage_ngs_pipe_cha(p,
                                        q,
                                        lam,
                                        va,
                                        D,
                                        S,
                                        dx,
                                        dt,
                                        M,
                                        pipe_name,
                                        idx_leak,
                                        d)
        case 'kt1':
            return leakage_ngs_pipe_kt1(p,
                                        q,
                                        lam,
                                        va,
                                        D,
                                        S,
                                        dx,
                                        M,
                                        pipe_name,
                                        idx_leak,
                                        d)

        case 'kt2':
            return leakage_ngs_pipe_kt2(p,
                                        q,
                                        lam,
                                        va,
                                        D,
                                        S,
                                        dx,
                                        M,
                                        pipe_name,
                                        idx_leak,
                                        d)
        case 'weno3':
            return leakage_ngs_pipe_weno3(p,
                                          q,
                                          lam,
                                          va,
                                          D,
                                          S,
                                          dx,
                                          M,
                                          pipe_name,
                                          idx_leak,
                                          d)
        case _:
            raise NotImplementedError(f'No such method: {method}!')



import numpy as np
from Solverz import Var, Param, Model, made_numerical, TimeSeriesParam, Rodas, Eqn, Opt, fdae_solver
from SolMuseum.pde import heat_pipe
import pandas as pd


def test_heat_pipe_iu(datadir):
    # modelling
    L = 9250
    dx = 370
    M = int(L / dx)
    Tinitial = np.zeros(M + 1)
    lam = 1 / 0.35
    Cp = 4182
    Ta = -10
    D = 1.4
    S = np.pi * (D / 2) ** 2
    rho = 960
    Tamb = -10

    for i in range(0, len(Tinitial)):
        fai = np.exp(-lam * i * dx / (Cp * np.abs(2543.5)))
        Tinitial[i] = 90.1724637976347 * fai + Ta * (1 - fai)

    df = pd.read_excel(datadir / 'param.xlsx',
                       sheet_name=None,
                       engine='openpyxl',
                       index_col=None
                       )
    mvariation = np.asarray(df['Sheet1']['m(m3/s)']) * rho
    Tmeasured = np.asarray(df['Sheet1']['Toutlet'])
    m = Model()
    m.T = Var('T', Tinitial)
    m.m = TimeSeriesParam('m',
                          mvariation,
                          np.linspace(0, 20 * 3600, 401))
    m.lam = Param('lam', lam)
    m.rho = Param('rho', rho)
    m.S = Param('S', S)
    m.Tamb = Param('Tamb', Tamb)
    m.Cp = Param('Cp', Cp)
    m.dt = Param('dt', 180)
    m.__dict__.update(heat_pipe(m.T,
                                m.m,
                                m.lam,
                                m.rho,
                                m.Cp,
                                m.S,
                                m.Tamb,
                                dx,
                                m.dt,
                                M,
                                '1',
                                method='iu'))
    m.Tsource = TimeSeriesParam('Tsource',
                                np.asarray(df['Sheet1']['Tsource']),
                                np.linspace(0, 20 * 3600, 401))
    m.Bd = Eqn('Bd', m.Tsource - m.T[0])
    heatpipe, y0 = m.create_instance()
    fdae = made_numerical(heatpipe, y0, sparse=True)

    sol = fdae_solver(fdae, [0, 20 * 3600], y0, Opt(step_size=180))

    res = pd.read_excel(datadir / 'res.xlsx',
                        sheet_name='Tout',
                        engine='openpyxl',
                        index_col=None
                        )
    bench_iu = res['iu']

    rmse = np.mean(np.abs(Tmeasured - sol.Y['T'][:, -1]) ** 2) ** (1 / 2)
    assert rmse < 0.2534
    np.testing.assert_allclose(bench_iu, sol.Y['T'][:, -1])


def test_heat_pipe_yao(datadir):
    # modelling
    L = 9250
    dx = 370
    M = int(L / dx)
    Tinitial = np.zeros(M + 1)
    lam = 1 / 0.35
    Cp = 4182
    Ta = -10
    D = 1.4
    S = np.pi * (D / 2) ** 2
    rho = 960
    Tamb = -10

    for i in range(0, len(Tinitial)):
        fai = np.exp(-lam * i * dx / (Cp * np.abs(2543.5)))
        Tinitial[i] = 90.1724637976347 * fai + Ta * (1 - fai)

    df = pd.read_excel(datadir / 'param.xlsx',
                       sheet_name=None,
                       engine='openpyxl',
                       index_col=None
                       )
    mvariation = np.asarray(df['Sheet1']['m(m3/s)']) * rho
    Tmeasured = np.asarray(df['Sheet1']['Toutlet'])
    m = Model()
    m.T = Var('T', Tinitial)
    m.m = TimeSeriesParam('m',
                          mvariation,
                          np.linspace(0, 20 * 3600, 401))
    m.lam = Param('lam', lam)
    m.rho = Param('rho', rho)
    m.S = Param('S', S)
    m.Tamb = Param('Tamb', Tamb)
    m.Cp = Param('Cp', Cp)
    m.dt = Param('dt', 180)
    m.__dict__.update(heat_pipe(m.T,
                                m.m,
                                m.lam,
                                m.rho,
                                m.Cp,
                                m.S,
                                m.Tamb,
                                dx,
                                m.dt,
                                M,
                                '1',
                                method='yao'))
    m.Tsource = TimeSeriesParam('Tsource',
                                np.asarray(df['Sheet1']['Tsource']),
                                np.linspace(0, 20 * 3600, 401))
    m.Bd = Eqn('Bd', m.Tsource - m.T[0])
    heatpipe, y0 = m.create_instance()
    fdae = made_numerical(heatpipe, y0, sparse=True)

    sol = fdae_solver(fdae, [0, 20 * 3600], y0, Opt(step_size=180))

    res = pd.read_excel(datadir / 'res.xlsx',
                        sheet_name='Tout',
                        engine='openpyxl',
                        index_col=None
                        )

    bench = res['yao']

    rmse = np.mean(np.abs(Tmeasured - sol.Y['T'][:, -1]) ** 2) ** (1 / 2)
    assert rmse < 0.1515
    np.testing.assert_allclose(bench, sol.Y['T'][:, -1])


def test_heat_pipe_rodaskt2(datadir):
    # modelling
    L = 9250
    dx = 370
    M = int(L / dx)
    Tinitial = np.zeros(M + 1)
    lam = 1 / 0.35
    Cp = 4182
    Ta = -10
    D = 1.4
    S = np.pi * (D / 2) ** 2
    rho = 960
    Tamb = -10

    for i in range(0, len(Tinitial)):
        fai = np.exp(-lam * i * dx / (Cp * np.abs(2543.5)))
        Tinitial[i] = 90.1724637976347 * fai + Ta * (1 - fai)

    df = pd.read_excel(datadir / 'param.xlsx',
                       sheet_name=None,
                       engine='openpyxl',
                       index_col=None
                       )
    mvariation = np.asarray(df['Sheet1']['m(m3/s)']) * rho
    Tmeasured = np.asarray(df['Sheet1']['Toutlet'])
    m = Model()
    m.T = Var('T', Tinitial)
    m.m = TimeSeriesParam('m',
                          mvariation,
                          np.linspace(0, 20 * 3600, 401))
    m.lam = Param('lam', lam)
    m.rho = Param('rho', rho)
    m.S = Param('S', S)
    m.Tamb = Param('Tamb', Tamb)
    m.Cp = Param('Cp', Cp)
    m.__dict__.update(heat_pipe(m.T,
                                m.m,
                                m.lam,
                                m.rho,
                                m.Cp,
                                m.S,
                                m.Tamb,
                                dx,
                                0,
                                M,
                                '1'))
    m.Tsource = TimeSeriesParam('Tsource',
                                np.asarray(df['Sheet1']['Tsource']),
                                np.linspace(0, 20 * 3600, 401))
    m.Bd = Eqn('Bd', m.Tsource - m.T[0])
    heatpipe, y0 = m.create_instance()
    dae = made_numerical(heatpipe, y0, sparse=True)

    dae.p['theta'] = np.array([2])
    soltheta2 = Rodas(dae, [0, 20 * 3600], y0, Opt(fix_h=True, hinit=180))

    dae.p['theta'] = np.array([1])
    soltheta1 = Rodas(dae, [0, 20 * 3600], y0, Opt(fix_h=True, hinit=180))

    res = pd.read_excel(datadir / 'res.xlsx',
                        sheet_name='Tout',
                        engine='openpyxl',
                        index_col=None
                        )
    bench_rodas_theta2 = res['rodastheta2']
    bench_rodas_theta1 = res['rodastheta1']

    rmsetheta2 = np.mean(np.abs(Tmeasured - soltheta2.Y['T'][:, -1]) ** 2) ** (1 / 2)
    rmsetheta1 = np.mean(np.abs(Tmeasured - soltheta1.Y['T'][:, -1]) ** 2) ** (1 / 2)
    assert rmsetheta2 < 0.0949
    assert rmsetheta1 < 0.1129
    np.testing.assert_allclose(bench_rodas_theta2, soltheta2.Y['T'][:, -1])
    np.testing.assert_allclose(bench_rodas_theta1, soltheta1.Y['T'][:, -1])




import numpy as np
from numba import njit


@njit(cache=True)
def minmod(a, b, c):
    if isinstance(a, (np.int32, np.int64, np.float64, np.float32, int, float)):
        a = np.array([a])
        a = a.reshape(-1)

    if isinstance(b, (np.int32, np.int64, np.float64, np.float32, int, float)):
        b = np.array([b])
        b = b.reshape(-1)

    if isinstance(c, (np.int32, np.int64, np.float64, np.float32, int, float)):
        c = np.array([c])
        c = c.reshape(-1)

    # check the consistency of input length
    if not (len(a) == len(b) == len(c)):
        raise ValueError("Input length must be the same!")

    res = np.zeros_like(a)

    for i in range(len(a)):
        if a[i] * b[i] > 0 and a[i] * c[i] > 0:
            res[i] = np.min(np.abs(np.array([a[i], b[i], c[i]]))) * np.sign(a[i])

    return res


@njit(cache=True)
def minmod_flag(a, b, c):
    """
    Return the index of minmod_flag
    """

    if isinstance(a, (np.int32, np.int64, np.float64, np.float32, int, float)):
        a = np.array([a])
        a = a.reshape(-1)

    if isinstance(b, (np.int32, np.int64, np.float64, np.float32, int, float)):
        b = np.array([b])
        b = b.reshape(-1)

    if isinstance(c, (np.int32, np.int64, np.float64, np.float32, int, float)):
        c = np.array([c])
        c = c.reshape(-1)

    # check the consistency of input length
    if not (len(a) == len(b) == len(c)):
        raise ValueError("Input length must be the same!")

    res = np.zeros_like(a).astype(np.int32)

    for i in range(len(a)):
        if a[i] * b[i] > 0 and a[i] * c[i] > 0:
            res[i] = np.abs(np.array([a[i], b[i], c[i]])).argmin() + 1

    return res


@njit(cache=True)
def switch_minmod(a, b, c, flag):
    """
    Conditionally output the derivatives of minmod according to the flag
    """

    if isinstance(a, (np.int32, np.int64, np.float64, np.float32, int, float)):
        a = np.array([a])
        a = a.reshape(-1)

    if isinstance(b, (np.int32, np.int64, np.float64, np.float32, int, float)):
        b = np.array([b])
        b = b.reshape(-1)

    if isinstance(c, (np.int32, np.int64, np.float64, np.float32, int, float)):
        c = np.array([c])
        c = c.reshape(-1)

    if isinstance(flag, (np.int32, np.int64, np.float64, np.float32, int, float)):
        flag = np.array([flag])
        flag = flag.reshape(-1)

    # if not (len(a) == len(b) == len(c) == len(flag)):
    #     raise ValueError("Input length must be the same!")

    res = np.zeros_like(flag)

    for i in range(len(flag)):
        if flag[i] == 1:
            if len(a) > 1:
                res[i] = a[i]
            else:
                res[i] = a[0]
        elif flag[i] == 2:
            if len(b) > 1:
                res[i] = b[i]
            else:
                res[i] = b[0]
        elif flag[i] == 3:
            if len(c) > 1:
                res[i] = c[i]
            else:
                res[i] = c[0]

    return res



from Solverz.sym_algebra.functions import MulVarFunc
from sympy import Integer


class SolPde(MulVarFunc):
    def _numpycode(self, printer, **kwargs):
        return (f'SolMF.pde.{self.__class__.__name__}' + r'(' +
                ', '.join([printer._print(arg, **kwargs) for arg in self.args]) + r')')


class minmod(SolPde):
    arglength = 3

    def _eval_derivative(self, s):
        return switch_minmod(*[arg.diff(s) for arg in self.args],
                             minmod_flag(*self.args))


class minmod_flag(SolPde):
    """
    Different from `minmod`, minmod function outputs the position of args instead of the values of args.
    """
    arglength = 3

    def _eval_derivative(self, s):
        return Integer(0)


class switch_minmod(SolPde):
    arglength = 4

    def _eval_derivative(self, s):
        return switch_minmod(*[arg.diff(s) for arg in self.args[0:len(self.args) - 1]], self.args[-1])



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
from datetime import datetime

project = "Solverz' Museum"
copyright = f'{datetime.now().year}, rzyu'
author = 'rzyu'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc', 'sphinx.ext.doctest',
              'sphinx_math_dollar', 'sphinx.ext.mathjax', 'numpydoc',
              'sphinx_reredirects', 'sphinx_copybutton',
              'sphinx.ext.graphviz', 'sphinxcontrib.jquery',
              'matplotlib.sphinxext.plot_directive', 'myst_parser',
              'sphinx.ext.intersphinx', ]  # 'sphinx.ext.linkcode'

# To stop docstrings inheritance.
autodoc_inherit_docstrings = False

# Sphinx是一个文档生成器，可以将Markdown或reStructuredText等文本格式转化为HTML、PDF等格式的文档。而MathJax是一个用于渲染数学公式的JavaScript库，它可以帮助将LaTeX或MathML格式的数学公式渲染为高质量的矢量图形。
# 虽然Sphinx本身提供了一些对数学公式的支持，但其渲染效果不如MathJax优秀。因此，为了获得更好的数学公式渲染效果，使用MathJax插件可以帮助Sphinx在生成文档时自动渲染数学公式，从而提高文档的质量和可读性。

templates_path = ['_templates']
exclude_patterns = []

source_suffix = {
    '.rst': 'restructuredtext',
    '.txt': 'markdown',
    '.md': 'markdown',
}

myst_enable_extensions = [
    "amsmath",
    "attrs_inline",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]

mathjax3_config = {
    "tex": {
        "inlineMath": [['\\(', '\\)']],
        "displayMath": [["\\[", "\\]"]],
        'packages': {'[+]': ['physics']}
    },
    'loader': {'load': ['[tex]/physics']},
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'furo'
html_title = "Solverz' Museum"
html_static_path = ['_static']
html_theme_options = {
    "description": "Distribute Your Muse Here!",
    "github_user": "smallbunnies",
    "github_repo": "SolMuseum",
    "fixed_sidebar": True
}



import numpy as np
from ..minmod_limiter import minmod, minmod_flag, switch_minmod


def test_minmod():
    a = np.array([-1, 1, 2, 4.])
    b = np.array([-3, 2, 0, 0])
    c = np.array([-5, 3, -1, 1.])
    np.testing.assert_allclose(minmod(a, b, c), np.array([-1, 1, 0, 0.]))


def test_minmod_flag():
    a = np.array([-1, 3, 2, -4, 5, 10])
    b = np.array([-3, 2, 1, -2, 0, -1])
    c = np.array([-5, 1, 1, -5.4, 1, 2])
    np.testing.assert_allclose(minmod_flag(a, b, c), np.array([1, 3, 2, 2, 0, 0]))


def test_switch_minmod():
    a = np.array([-1, 3, 2, -4, 5, 10])
    b = np.array([-3, 2, 1, -2, 0, -1])
    c = np.array([-5, 1, 1, -5.4, 1, 2])
    flag = minmod_flag(a, b, c)
    np.testing.assert_allclose(switch_minmod(a, b, c, flag), np.array([-1, 1, 1, -2, 0, 0.0]))



import numpy as np
from Solverz import Model, Param, Var, made_numerical, Rodas, Opt, Eqn, TimeSeriesParam
import pandas as pd
from SolMuseum.pde.gas.broken_pipe import leakage_pipe


# %%
def test_kt2(shared_datadir):
    # %%
    L = 51000
    T = 2000
    lam = 0.03
    dx = 500
    M = int(L / dx)
    idx_leak = int(M * 0.5)
    p0 = 6621246.69079594
    q0 = 14

    m = Model()
    m.D = Param('D', value=0.5901)
    S = np.pi * (m.D / 2) ** 2
    m.lam = Param('lam', lam)
    m.va = Param('va', value=340)
    m.p = Var('p', value=p0 * np.ones((M + 1)))
    m.q = Var('q', value=q0 * np.ones((M + 1)))

    m.__dict__.update(leakage_pipe(m.p,
                                   m.q,
                                   m.lam,
                                   m.va,
                                   m.D,
                                   S,
                                   dx,
                                   0,
                                   M,
                                   '1',
                                   idx_leak,
                                   0.9 * m.D,
                                   method='kt2'))
    m.theta = Param('theta', 2)
    m.pb = Param('pb', p0)
    m.qb = Param('qb', q0)
    m.bd1 = Eqn('bd1', m.p[0] - m.pb)
    m.bd2 = Eqn('bd2', m.q[M] - m.qb)

    # %% solution
    sdae, y0 = m.create_instance()
    ndae, code = made_numerical(sdae, y0, output_code=True, sparse=True)

    ndae.p['leak_rate'] = TimeSeriesParam('leak_rate',
                                          v_series=[0, 0, 1, 1],
                                          time_series=[0, 100, 105, 3600])

    sol = Rodas(ndae, np.linspace(0, T, 401), y0, Opt(pbar=True))

    # %%
    # import matplotlib.pyplot as plt
    #
    # plt.plot(sol.T, sol.Y['p'][:, 0])
    # plt.plot(sol.T, sol.Y['p'][:, -1])
    # plt.plot(sol.T, sol.Y['p'][:, idx_leak])
    # plt.plot(sol.T, sol.Y['p'][:, idx_leak - 1])
    # plt.show()
    #
    # plt.plot(sol.T, sol.Y['q'][:, 0])
    # # plt.plot(sol.T, sol.Y['q_1_leak1'])
    # plt.plot(sol.T, sol.Y['q'][:, -1])
    # plt.show()
    #
    # res = dict()
    # res['pout'] = sol.Y['p'][:, -1]
    # res['qin'] = sol.Y['q'][:, 0]
    # res['qupstream'] = sol.Y['q_1_leak1'].reshape(-1)
    # df = pd.DataFrame(res)
    #
    # with pd.ExcelWriter(f'data/res.xlsx', engine='openpyxl', mode='a', if_sheet_exists='replace') as writer:
    #     # Write each DataFrame to a different sheet
    #     df.to_excel(writer, sheet_name='kt2')

    df = pd.read_excel(shared_datadir / 'res.xlsx',
                       sheet_name='kt2',
                       engine='openpyxl',
                       index_col=None
                       )
    qin = np.asarray(df['qin'])
    pout = np.asarray(df['pout'])
    qupstream = np.asarray(df['qupstream'])
    np.testing.assert_allclose(qin, sol.Y['q'][:, 0])
    np.testing.assert_allclose(pout, sol.Y['p'][:, -1])
    np.testing.assert_allclose(qupstream, sol.Y['q_1_leak1'].reshape(-1))



# Pls do not import any symbolic function or module here.
# If doing so, Solverz cannot import num_api module independently from SolMuseum because the symbolic functions and
# modules have to be initialized in this __init__.py file where the symbols are imported from Solverz.
# This causes the cirluar ImportError.
# This is ad-hoc. We are working on this for a more elegant solution.
